from edsl.exceptions.BaseException import BaseException


# from edsl.utilities.utilities import is_notebook

# from IPython.core.error import UsageError

# class AgentListErrorAlternative(UsageError):
#     def __init__(self, message):
#         super().__init__(message)

import sys
from edsl.utilities.is_notebook import is_notebook


class AgentListError(BaseException):
    relevant_doc = "https://docs.expectedparrot.com/en/latest/agents.html#agent-lists"


class AgentErrors(BaseException):
    relevant_doc = "https://docs.expectedparrot.com/en/latest/agents.html"


class AgentDynamicTraitsFunctionError(AgentErrors):
    relevant_doc = (
        "https://docs.expectedparrot.com/en/latest/agents.html#dynamic-traits-function"
    )
    relevant_notebook = "https://docs.expectedparrot.com/en/latest/notebooks/example_agent_dynamic_traits.html"


class AgentDirectAnswerFunctionError(AgentErrors):
    relevant_doc = "https://docs.expectedparrot.com/en/latest/agents.html#agent-direct-answering-methods"


class AgentCombinationError(AgentErrors):
    relevant_doc = (
        "https://docs.expectedparrot.com/en/latest/agents.html#combining-agents"
    )


class AgentNameError(AgentErrors):
    relevant_doc = "https://docs.expectedparrot.com/en/latest/agents.html#agent-names"


class AgentTraitKeyError(AgentErrors):
    relevant_doc = (
        "https://docs.expectedparrot.com/en/latest/agents.html#constructing-an-agent"
    )


class FailedTaskException(Exception):
    def __init__(self, message, agent_response_dict):
        super().__init__(message)
        self.agent_response_dict = agent_response_dict



from edsl.scenarios.file_methods import FileMethods
import tempfile


class TxtMethods(FileMethods):
    suffix = "txt"

    def view_system(self):
        import os
        import subprocess

        if os.path.exists(self.path):
            try:
                if (os_name := os.name) == "posix":
                    subprocess.run(["open", self.path], check=True)  # macOS
                elif os_name == "nt":
                    os.startfile(self.path)  # Windows
                else:
                    subprocess.run(["xdg-open", self.path], check=True)  # Linux
            except Exception as e:
                print(f"Error opening TXT: {e}")
        else:
            print("TXT file was not found.")

    def view_notebook(self):
        from IPython.display import FileLink, display

        display(FileLink(self.path))

    def example(self):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".txt") as f:
            f.write(b"Hello, World!")
        return f.name



from __future__ import annotations
from typing import Optional, Any, List, Annotated, Literal

from pydantic import BaseModel, Field

from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import (
    QuestionOptionsDescriptor,
    NumSelectionsDescriptor,
)
from edsl.questions.response_validator_abc import ResponseValidatorABC


def create_response_model(
    choices: list,
    num_selections: Optional[int] = None,
    permissive: bool = False,
):
    """
    :param choices: A list of allowed values for the answer field.
    :param include_comment: Whether to include a comment field in the model.
    :return: A new Pydantic model class.
    """
    # Convert the choices list to a tuple for use with Literal
    choice_tuple = tuple(choices)

    field_params = {}
    if num_selections is not None and not permissive:
        field_params["min_items"] = num_selections
        field_params["max_items"] = num_selections

    class RankResponse(BaseModel):
        answer: Annotated[
            List[Literal[choice_tuple]],
            Field(..., **field_params),
        ] = Field(..., description="List of selected choices")
        comment: Optional[str] = Field(None, description="Optional comment field")
        generated_tokens: Optional[Any] = Field(None)

        class Config:
            @staticmethod
            def json_schema_extra(schema: dict, model: BaseModel) -> None:
                # Add the list of choices to the schema for better documentation
                for prop in schema.get("properties", {}).values():
                    if prop.get("title") == "answer":
                        prop["items"] = {"enum": choices}

    return RankResponse


class RankResponseValidator(ResponseValidatorABC):
    required_params = ["num_selections", "permissive", "use_code", "question_options"]
    valid_examples = []
    invalid_examples = []

    def fix(self, response, verbose=False):
        if verbose:
            print("Invalid response of QuestionRank was: ", False)
        response_text = response.get("generated_tokens")
        if response_text is None or response_text == "":  # nothing to be done
            return response
        # Maybe it's a comma separated list?
        response_text = str(response.get("answer"))
        proposed_list = (
            response_text.replace("[", "").replace("]", "").replace("'", "").split(",")
        )
        proposed_list = [item.strip() for item in proposed_list]

        if verbose:
            print("Using code? ", self.use_code)
        if self.use_code:
            try:
                proposed_list = [int(i) for i in proposed_list]
            except ValueError:
                # print("Could not convert to int")
                pass

        if verbose:
            print("Proposed solution is: ", proposed_list)

        # print(f"Ivalid generated tokens was was: {response_text}")
        if "comment" in response:
            proposed_data = {
                "answer": proposed_list,
                "comment": response["comment"],
                "generated_tokens": response.get("generated_tokens", None),
            }
        else:
            proposed_data = {
                "answer": proposed_list,
                "generated_tokens": response.get("generated_tokens", None),
            }

        try:
            self.response_model(**proposed_data)
            return proposed_data
        except Exception as e:
            if verbose:
                print(f"Proposed solution {proposed_data} is invalid. Error: {e}")
            # return response
        if verbose:
            print("Now seeing if responses show up in the answer")
        matches = []
        for index, option in enumerate(self.question_options):
            if self.use_code:
                if str(index) in response_text:
                    if index not in matches:
                        matches.append(index)
            else:
                if option in response_text:
                    if option not in matches:
                        matches.append(option)
        proposed_data = {
            "answer": matches,
            "comment": response.get("comment", None),
            "generated_tokens": response.get("generated_tokens", None),
        }
        try:
            self.response_model(**proposed_data)
            return proposed_data
        except Exception as e:
            if verbose:
                print(f"Proposed solution {proposed_data} is invalid. Error: {e}")
            return response


class QuestionRank(QuestionBase):
    """This question prompts the agent to rank options from a list."""

    question_type = "rank"
    question_options: list[str] = QuestionOptionsDescriptor()
    num_selections = NumSelectionsDescriptor()

    _response_model = None
    response_validator_class = RankResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        num_selections: Optional[int] = None,
        question_presentation: Optional[str] = None,
        answering_instructions: Optional[str] = None,
        permissive: bool = False,
        use_code: bool = True,
        include_comment: bool = True,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param question_options: The options the respondent should select from.
        :param min_selections: The minimum number of options that must be selected.
        :param max_selections: The maximum number of options that must be selected.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.question_options = question_options
        self.num_selections = num_selections or len(question_options)
        self.question_presentation = question_presentation
        self.answering_instructions = answering_instructions
        self.permissive = permissive
        self.use_code = use_code
        self.include_comment = include_comment

    def create_response_model(self):
        choices = (
            self.question_options
            if not self.use_code
            else range(len(self.question_options))
        )
        return create_response_model(
            choices=choices,
            num_selections=self.num_selections,
            permissive=self.permissive,
        )

    ################
    # Answer methods
    ################
    # def _validate_answer(self, answer: Any) -> dict[str, list[int]]:
    #     """Validate the answer."""
    #     self._validate_answer_template_basic(answer)
    #     self._validate_answer_key_value(answer, "answer", list)
    #     self._validate_answer_rank(answer)
    #     return answer

    def _translate_answer_code_to_answer(
        self, answer_codes, scenario: Scenario = None
    ) -> list[str]:
        """Translate the answer code to the actual answer."""
        from edsl.scenarios.Scenario import Scenario
        from jinja2 import Template

        scenario = scenario or Scenario()
        translated_options = [
            Template(option).render(scenario) for option in self.question_options
        ]
        translated_codes = []
        for answer_code in answer_codes:
            if self._use_code:
                translated_codes.append(translated_options[int(answer_code)])
            else:
                translated_codes.append(answer_code)
        return translated_codes

    # def _simulate_answer(self, human_readable=True) -> dict[str, Union[int, str]]:
    #     """Simulate a valid answer for debugging purposes."""
    #     from edsl.utilities.utilities import random_string

    #     if human_readable:
    #         selected = random.sample(self.question_options, self.num_selections)
    #     else:
    #         selected = random.sample(
    #             range(len(self.question_options)), self.num_selections
    #         )
    #     answer = {
    #         "answer": selected,
    #         "comment": random_string(),
    #     }
    #     return answer

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        <form id="rankForm">
            <p>{{ question_text }}</p>
            {% for option in question_options %}
            <div>
                <label for="{{ option }}">{{ option }}</label>
                <input type="number" id="{{ option }}" name="{{ question_name }}[{{ option }}]" value="0" min="1" max="{{ question_options|length }}" oninput="updateRankings()">
            </div>
            {% endfor %}
        </form>
        <script>
        function updateRankings() {
            let options = {{ question_options|length }};
            let values = [];
            let isValid = true;

            {% for option in question_options %}
            let value = parseInt(document.getElementById("{{ option }}").value) || 0;
            if (value > 0 && value <= options && !values.includes(value)) {
                values.push(value);
            } else if (value !== 0) {
                isValid = false;
            }
            {% endfor %}

            if (!isValid || values.length !== new Set(values).size) {
                document.getElementById("error").innerText = "Please enter unique and valid ranks for each option.";
            } else {
                document.getElementById("error").innerText = "";
            }
        }
        </script>
        <p id="error" style="color: red;"></p>
        """
        ).render(
            question_name=self.question_name,
            question_text=self.question_text,
            question_options=self.question_options,
        )
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls, use_code=False, include_comment=True) -> QuestionRank:
        """Return an example question."""
        return cls(
            question_name="rank_foods",
            question_text="Rank your favorite foods.",
            question_options=["Pizza", "Pasta", "Salad", "Soup"],
            num_selections=2,
            use_code=use_code,
            include_comment=include_comment,
        )


def main():
    """Show example usage."""
    from edsl.questions.QuestionRank import QuestionRank

    q = QuestionRank.example(use_code=True)
    q.question_text
    q.question_name
    q.question_options
    q.num_selections
    # validate an answer
    answer = {"answer": [0, 1], "comment": "I like pizza and pasta."}
    q._validate_answer(answer)
    # translate an answer code to an answer
    # q._translate_answer_code_to_answer([0, 1])
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    q = QuestionRank.example(use_code=False)
    answer = {"answer": ["Pizza", "Pasta"], "comment": "I like pizza and pasta."}
    q._validate_answer(answer)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from typing import Union, Optional
import asyncio
import time
import aiohttp


class TokenBucketClient:
    """REST API client version of TokenBucket that maintains the same interface
    by delegating to a server running the original TokenBucket implementation."""

    def __init__(
        self,
        *,
        bucket_name: str,
        bucket_type: str,
        capacity: Union[int, float],
        refill_rate: Union[int, float],
        api_base_url: str = "http://localhost:8000",
    ):
        self.bucket_name = bucket_name
        self.bucket_type = bucket_type
        self.capacity = capacity
        self.refill_rate = refill_rate
        self.api_base_url = api_base_url
        self.bucket_id = f"{bucket_name}_{bucket_type}"

        # Initialize the bucket on the server
        asyncio.run(self._create_bucket())

        # Cache some values locally
        self.creation_time = time.monotonic()
        self.turbo_mode = False

    async def _create_bucket(self):
        async with aiohttp.ClientSession() as session:
            payload = {
                "bucket_name": self.bucket_name,
                "bucket_type": self.bucket_type,
                "capacity": self.capacity,
                "refill_rate": self.refill_rate,
            }
            async with session.post(
                f"{self.api_base_url}/bucket",
                json=payload,
            ) as response:
                if response.status != 200:
                    raise ValueError(f"Unexpected error: {await response.text()}")

                result = await response.json()
                if result["status"] == "existing":
                    # Update our local values to match the existing bucket
                    self.capacity = float(result["bucket"]["capacity"])
                    self.refill_rate = float(result["bucket"]["refill_rate"])

    def turbo_mode_on(self):
        """Set the refill rate to infinity."""
        asyncio.run(self._set_turbo_mode(True))
        self.turbo_mode = True

    def turbo_mode_off(self):
        """Restore the refill rate to its original value."""
        asyncio.run(self._set_turbo_mode(False))
        self.turbo_mode = False

    async def add_tokens(self, amount: Union[int, float]):
        """Add tokens to the bucket."""
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.api_base_url}/bucket/{self.bucket_id}/add_tokens",
                params={"amount": amount},
            ) as response:
                if response.status != 200:
                    raise ValueError(f"Failed to add tokens: {await response.text()}")

    async def _set_turbo_mode(self, state: bool):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.api_base_url}/bucket/{self.bucket_id}/turbo_mode/{str(state).lower()}"
            ) as response:
                if response.status != 200:
                    raise ValueError(
                        f"Failed to set turbo mode: {await response.text()}"
                    )

    async def get_tokens(
        self, amount: Union[int, float] = 1, cheat_bucket_capacity=True
    ) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.api_base_url}/bucket/{self.bucket_id}/get_tokens",
                params={
                    "amount": amount,
                    "cheat_bucket_capacity": int(cheat_bucket_capacity),
                },
            ) as response:
                if response.status != 200:
                    raise ValueError(f"Failed to get tokens: {await response.text()}")

    def get_throughput(self, time_window: Optional[float] = None) -> float:
        status = asyncio.run(self._get_status())
        now = time.monotonic()

        if time_window is None:
            start_time = self.creation_time
        else:
            start_time = now - time_window

        if start_time < self.creation_time:
            start_time = self.creation_time

        elapsed_time = now - start_time

        if elapsed_time == 0:
            return status["num_released"] / 0.001

        return (status["num_released"] / elapsed_time) * 60

    async def _get_status(self) -> dict:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"{self.api_base_url}/bucket/{self.bucket_id}/status"
            ) as response:
                if response.status != 200:
                    raise ValueError(
                        f"Failed to get bucket status: {await response.text()}"
                    )
                return await response.json()

    def __add__(self, other) -> "TokenBucketClient":
        """Combine two token buckets."""
        return TokenBucketClient(
            bucket_name=self.bucket_name,
            bucket_type=self.bucket_type,
            capacity=min(self.capacity, other.capacity),
            refill_rate=min(self.refill_rate, other.refill_rate),
            api_base_url=self.api_base_url,
        )

    @property
    def tokens(self) -> float:
        """Get the number of tokens remaining in the bucket."""
        status = asyncio.run(self._get_status())
        return float(status["tokens"])

    def wait_time(self, requested_tokens: Union[float, int]) -> float:
        """Calculate the time to wait for the requested number of tokens."""
        # self.refill()  # Update the current token count
        if self.tokens >= float(requested_tokens):
            return 0.0
        try:
            return (requested_tokens - self.tokens) / self.refill_rate
        except Exception as e:
            raise ValueError(f"Error calculating wait time: {e}")

    # def wait_time(self, num_tokens: Union[int, float]) -> float:
    #     return 0  # TODO - Need to implement this on the server side

    def visualize(self):
        """Visualize the token bucket over time."""
        status = asyncio.run(self._get_status())
        times, tokens = zip(*status["log"])
        start_time = times[0]
        times = [t - start_time for t in times]

        from matplotlib import pyplot as plt

        plt.figure(figsize=(10, 6))
        plt.plot(times, tokens, label="Tokens Available")
        plt.xlabel("Time (seconds)", fontsize=12)
        plt.ylabel("Number of Tokens", fontsize=12)
        details = f"{self.bucket_name} ({self.bucket_type}) Bucket Usage Over Time\nCapacity: {self.capacity:.1f}, Refill Rate: {self.refill_rate:.1f}/second"
        plt.title(details, fontsize=14)
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # bucket = TokenBucketClient(
    #     bucket_name="test", bucket_type="test", capacity=100, refill_rate=10
    # )
    # asyncio.run(bucket.get_tokens(50))
    # time.sleep(1)  # Wait for 1 second
    # asyncio.run(bucket.get_tokens(30))
    # throughput = bucket.get_throughput(1)
    # print(throughput)
    # bucket.visualize()



from __future__ import annotations
import random
from typing import Any, Optional, Union

from jinja2 import Template

from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import (
    IntegerDescriptor,
    QuestionOptionsDescriptor,
)

from edsl.questions.decorators import inject_exception

from pydantic import field_validator
from edsl.questions.response_validator_abc import ResponseValidatorABC
from edsl.questions.data_structures import BaseResponse

from edsl.exceptions.questions import QuestionAnswerValidationError

from pydantic import BaseModel, Field, conlist
from typing import List, Literal, Optional, Annotated


def create_checkbox_response_model(
    choices: list,
    min_selections: Optional[int] = None,
    max_selections: Optional[int] = None,
    permissive: bool = False,
):
    """
    Dynamically create a CheckboxResponse model with a predefined list of choices.

    :param choices: A list of allowed values for the answer field.
    :param include_comment: Whether to include a comment field in the model.
    :return: A new Pydantic model class.
    """
    # Convert the choices list to a tuple for use with Literal
    choice_tuple = tuple(choices)

    field_params = {}
    if min_selections is not None and not permissive:
        field_params["min_items"] = min_selections
    if max_selections is not None and not permissive:
        field_params["max_items"] = max_selections

    class CheckboxResponse(BaseModel):
        answer: Annotated[
            List[Literal[choice_tuple]],
            Field(..., **field_params),
        ] = Field(..., description="List of selected choices")
        comment: Optional[str] = Field(None, description="Optional comment field")
        generated_tokens: Optional[Any] = Field(default=None)

        class Config:
            @staticmethod
            def json_schema_extra(schema: dict, model: BaseModel) -> None:
                # Add the list of choices to the schema for better documentation
                for prop in schema.get("properties", {}).values():
                    if prop.get("title") == "answer":
                        prop["items"] = {"enum": choices}

    return CheckboxResponse


class CheckBoxResponseValidator(ResponseValidatorABC):
    required_params = [
        "question_options",
        "min_selections",
        "max_selections",
        "use_code",
        "permissive",
    ]

    valid_examples = [
        ({"answer": [1, 2]}, {"question_options": ["Good", "Great", "OK", "Bad"]})
    ]

    invalid_examples = [
        (
            {"answer": [-1]},
            {"question_options": ["Good", "Great", "OK", "Bad"]},
            "Answer code must be a non-negative integer",
        ),
        (
            {"answer": 1},
            {"question_options": ["Good", "Great", "OK", "Bad"]},
            "Answer code must be a list",
        ),
        (
            {"answer": [1, 2, 3, 4]},
            {
                "question_options": ["Good", "Great", "OK", "Bad"],
                "min_selections": 1,
                "max_selections": 2,
            },
            "Too many options selected",
        ),
    ]

    def fix(self, response, verbose=False):
        if verbose:
            print("Invalid response of QuestionCheckBox was: ", response)
        response_text = response.get("generated_tokens")
        if response_text is None or response_text == "":  # nothing to be done
            return response
        # Maybe it's a comma separated list?
        proposed_list = response_text.split(",")
        proposed_list = [item.strip() for item in proposed_list]
        if verbose:
            print("Using code? ", self.use_code)
        if self.use_code:
            try:
                proposed_list = [int(i) for i in proposed_list]
            except ValueError:
                # print("Could not convert to int")
                pass

        if verbose:
            print("Proposed solution is: ", proposed_list)

        # print(f"Ivalid generated tokens was was: {response_text}")
        if "comment" in response:
            proposed_data = {
                "answer": proposed_list,
                "comment": response["comment"],
                "generated_tokens": response.get("generated_tokens", None),
            }
        else:
            proposed_data = {
                "answer": proposed_list,
                "generated_tokens": response.get("generated_tokens", None),
            }

        try:
            self.response_model(**proposed_data)
            print("Proposed solution is valid")
            print("Returning proposed data: ", proposed_data)
            return proposed_data
        except Exception as e:
            if verbose:
                print(f"Proposed solution {proposed_data} is invalid. Error: {e}")
            # return response
        if verbose:
            print("Now seeing if responses show up in the answer")
        matches = []
        for index, option in enumerate(self.question_options):
            if self.use_code:
                if str(index) in response_text:
                    matches.append(index)
            else:
                if option in response_text:
                    matches.append(index)
        proposed_data = {
            "answer": matches,
            "comment": response.get("comment", None),
            "generated_tokens": response.get("generated_tokens", None),
        }
        try:
            self.response_model(**proposed_data)
            return proposed_data
        except Exception as e:
            if verbose:
                print(f"Proposed solution {proposed_data} is invalid. Error: {e}")
            return response

    def custom_validate(self, response) -> BaseResponse:
        if response.answer is None:
            raise QuestionAnswerValidationError("Answer is missing.")
        return response.dict()


class QuestionCheckBox(QuestionBase):
    """This question prompts the agent to select options from a list."""

    question_type = "checkbox"
    purpose = "When options are known and limited"
    question_options: list[str] = QuestionOptionsDescriptor()
    min_selections = IntegerDescriptor(none_allowed=True)
    max_selections = IntegerDescriptor(none_allowed=True)

    _response_model = None
    response_validator_class = CheckBoxResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        min_selections: Optional[int] = None,
        max_selections: Optional[int] = None,
        include_comment: bool = True,
        use_code: bool = True,
        question_presentation: Optional[str] = None,
        answering_instructions: Optional[str] = None,
        permissive: bool = False,
    ):
        """Instantiate a new QuestionCheckBox.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param question_options: The options the respondent should select from.
        :param min_selections: The minimum number of options that must be selected.
        :param max_selections: The maximum number of options that must be selected.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.min_selections = min_selections
        self.max_selections = max_selections
        self.question_options = question_options

        self._include_comment = include_comment
        self._use_code = use_code
        self.permissive = permissive

        self.question_presentation = question_presentation
        self.answering_instructions = answering_instructions

    def create_response_model(self):
        if not self._use_code:
            return create_checkbox_response_model(
                self.question_options,
                min_selections=self.min_selections,
                max_selections=self.max_selections,  # include_comment=self._include_comment
                permissive=self.permissive,
            )
        else:
            return create_checkbox_response_model(
                list(range(len(self.question_options))),
                min_selections=self.min_selections,
                max_selections=self.max_selections,  # include_comment=self._include_comment
                permissive=self.permissive,
            )

    def _translate_answer_code_to_answer(
        self, answer_codes, scenario: "Scenario" = None
    ):
        """
        Translate the answer code to the actual answer.

        For example, for question options ["a", "b", "c"],the answer codes are 0, 1, and 2.
        The LLM will respond with [0,1] and this code will translate it to ["a","b"].
        """
        from edsl.scenarios.Scenario import Scenario

        scenario = scenario or Scenario()
        translated_options = [
            Template(str(option)).render(scenario) for option in self.question_options
        ]
        translated_codes = []
        for answer_code in answer_codes:
            if self._use_code:
                translated_codes.append(translated_options[int(answer_code)])
            else:
                translated_codes.append(answer_code)
        return translated_codes

    # def _simulate_answer(self, human_readable=True) -> dict[str, Union[int, str]]:
    #     """Simulate a valid answer for debugging purposes."""
    #     from edsl.utilities.utilities import random_string

    #     min_selections = self.min_selections or 1
    #     max_selections = self.max_selections or len(self.question_options)
    #     num_selections = random.randint(min_selections, max_selections)
    #     if human_readable:
    #         # Select a random number of options from self.question_options
    #         selected_options = random.sample(self.question_options, num_selections)
    #         answer = {
    #             "answer": selected_options,
    #             "comment": random_string(),
    #         }
    #     else:
    #         # Select a random number of indices from the range of self.question_options
    #         selected_indices = random.sample(
    #             range(len(self.question_options)), num_selections
    #         )
    #         answer = {
    #             "answer": selected_indices,
    #             "comment": random_string(),
    #         }
    #     return answer

    @property
    def question_html_content(self) -> str:
        instructions = ""
        if self.min_selections is not None:
            instructions += f"Select at least {self.min_selections} option(s). "
        if self.max_selections is not None:
            instructions += f"Select at most {self.max_selections} option(s)."
        question_html_content = Template(
            """
        <p>{{ instructions }}</p>
        {% for option in question_options %} 
        <div>
        <input type="checkbox" id="{{ option }}" name="{{ question_name }}" value="{{ option }}">
        <label for="{{ option }}">{{ option }}</label>
        </div>
        {% endfor %}
        """
        ).render(
            instructions=instructions,
            question_name=self.question_name,
            question_options=self.question_options,
        )
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    @inject_exception
    def example(cls, include_comment=False, use_code=True) -> QuestionCheckBox:
        """Return an example checkbox question."""
        return cls(
            question_name="never_eat",
            question_text="Which of the following foods would you eat if you had to?",
            question_options=[
                "soggy meatpie",
                "rare snails",
                "mouldy bread",
                "panda milk custard",
                "McDonalds",
            ],
            min_selections=2,
            max_selections=5,
            use_code=use_code,
            include_comment=include_comment,
        )


def main():
    """Create an example QuestionCheckBox and test its methods."""
    from edsl.questions.QuestionCheckBox import QuestionCheckBox

    q = QuestionCheckBox.example()
    q.question_text
    q.question_options
    q.question_name
    # validate an answer
    q._validate_answer({"answer": [1, 2], "comment": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer([1, 2])
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""This module contains the descriptors used to validate the attributes of the question classes."""

from abc import ABC, abstractmethod
import re
import textwrap
from typing import Any, Callable, List, Optional
from edsl.exceptions.questions import (
    QuestionCreationValidationError,
    QuestionAnswerValidationError,
)
from edsl.questions.settings import Settings


################################
# Helper functions
################################


def contains_single_braced_substring(s: str) -> bool:
    """Check if the string contains a substring in single braces."""
    pattern = r"(?<!\{)\{[^{}]+\}(?!\})"
    match = re.search(pattern, s)
    return bool(match)


def is_number(value: Any) -> bool:
    """Check if an object is a number."""
    return isinstance(value, int) or isinstance(value, float)


def is_number_or_none(value: Any) -> bool:
    """Check if an object is a number or None."""
    return value is None or is_number(value)


################################
# Descriptor ABC
################################


class BaseDescriptor(ABC):
    """ABC for something."""

    @abstractmethod
    def validate(self, value: Any) -> None:
        """Validate the value. If it is invalid, raises an exception. If it is valid, does nothing."""
        pass

    def __get__(self, instance, owner):
        """Get the value of the attribute."""
        if self.name not in instance.__dict__:
            return {}
        return instance.__dict__[self.name]

    def __set__(self, instance, value: Any) -> None:
        """Set the value of the attribute."""
        new_value = self.validate(value, instance)

        if new_value is not None:
            instance.__dict__[self.name] = new_value
        else:
            instance.__dict__[self.name] = value

    def __set_name__(self, owner, name: str) -> None:
        """Set the name of the attribute."""
        self.name = "_" + name


################################
# General descriptors
################################


class FunctionDescriptor(BaseDescriptor):
    """Validate that a value is a function."""

    def validate(self, value: Any, instance) -> Callable:
        """Validate the value is a function, and if so, returns it."""
        if not callable(value):
            raise QuestionCreationValidationError(
                f"Expected a function (got {value}).)"
            )
        return value


class IntegerDescriptor(BaseDescriptor):
    """
    Validate that a value is an integer.

    - `none_allowed` is whether None is allowed as a value.
    """

    def __init__(self, none_allowed: bool = False):
        """Initialize the descriptor."""
        self.none_allowed = none_allowed

    def validate(self, value, instance):
        """Validate the value is an integer."""
        if self.none_allowed:
            if not (isinstance(value, int) or value is None):
                raise QuestionAnswerValidationError(
                    f"Expected an integer or None (got {value})."
                )
        else:
            if not isinstance(value, int):
                raise QuestionAnswerValidationError(
                    f"Expected an integer (got {value})."
                )


class IntegerOrNoneDescriptor(BaseDescriptor):
    """Validate that a value is an integer or None."""

    def validate(self, value, instance):
        """Validate the value is an integer or None."""
        if not (isinstance(value, int) or value is None):
            raise QuestionCreationValidationError(
                f"Expected an integer or None (got {value})."
            )


class NumericalOrNoneDescriptor(BaseDescriptor):
    """Validate that a value is a number or None."""

    def validate(self, value, instance):
        """Validate the value is a number or None."""
        if not is_number_or_none(value):
            raise QuestionAnswerValidationError(
                f"Expected a number or None (got {value})."
            )


################################
# Attribute-specific descriptors
################################


class AnswerTemplateDescriptor(BaseDescriptor):
    """Validate that the answer template is a dictionary with string keys and string values."""

    def validate(self, value: Any, instance) -> None:
        """Validate the answer template."""
        if not isinstance(value, dict):
            raise QuestionCreationValidationError(
                f"`answer_template` must be a dictionary (got {value}).)"
            )
        if not all(isinstance(x, str) for x in value.keys()):
            raise QuestionCreationValidationError(
                f"`answer_template` keys must be strings (got {value})."
            )


class InstructionsDescriptor(BaseDescriptor):
    """Validate that the `instructions` attribute is a string."""

    def validate(self, value, instance):
        """Validate the value is a string."""
        # if not isinstance(value, str):
        #     raise QuestionCreationValidationError(
        #         f"Question `instructions` must be a string (got {value})."
        #     )
        pass


class NumSelectionsDescriptor(BaseDescriptor):
    """Validate that `num_selections` is an integer, is less than the number of options, and is positive."""

    def validate(self, value, instance):
        """Validate the value is an integer, is less than the number of options, and is positive."""
        if not (isinstance(value, int)):
            raise QuestionCreationValidationError(
                f"`num_selections` must be an integer (got {value})."
            )
        if value > len(instance.question_options):
            raise QuestionAnswerValidationError(
                f"`num_selections` must be less than the number of options (got {value})."
            )
        if value < 1:
            raise QuestionAnswerValidationError(
                f"`num_selections` must a positive integer (got {value})."
            )


class OptionLabelDescriptor(BaseDescriptor):
    """Validate that the `option_label` attribute is a string.

    >>> class TestQuestion:
    ...     option_label = OptionLabelDescriptor()
    ...     def __init__(self, option_label: str):
    ...         self.option_label = option_label

    >>> _ = TestQuestion("{{Option}}")

    """

    def validate(self, value, instance):
        """Validate the value is a string."""
        if isinstance(value, str):
            if "{{" in value and "}}" in value:
                # they're trying to use a dynamic question name - let's let this play out
                return None

        key_values = [int(v) for v in value.keys()]

        if value and (key_values := [float(v) for v in value.keys()]) != []:
            if min(key_values) != min(instance.question_options):
                raise QuestionCreationValidationError(
                    f"First option needs a label (got {value})"
                )
            if max(key_values) != max(instance.question_options):
                raise QuestionCreationValidationError(
                    f"Last option needs a label (got {value})"
                )
            if not all(isinstance(x, str) for x in value.values()):
                raise QuestionCreationValidationError(
                    "Option labels must be strings (got {value})."
                )
            for key in key_values:
                if key not in instance.question_options:
                    raise QuestionCreationValidationError(
                        f"Option label key ({key}) is not in question options ({instance.question_options})."
                    )

            if len(value.values()) != len(set(value.values())):
                raise QuestionCreationValidationError(
                    f"Option labels must be unique (got {value})."
                )


class QuestionNameDescriptor(BaseDescriptor):
    """Validate that the `question_name` attribute is a valid variable name."""

    def validate(self, value, instance):
        """Validate the value is a valid variable name."""
        from edsl.utilities.utilities import is_valid_variable_name

        if "{{" in value and "}}" in value:
            # they're trying to use a dynamic question name - let's let this play out
            return None

        if value.endswith("_comment") or value.endswith("_generated_tokens"):
            raise QuestionCreationValidationError(
                f"`question_name` cannot end with '_comment' or '_generated_tokens - (got {value})."
            )

        if not is_valid_variable_name(value):
            raise QuestionCreationValidationError(
                f"`question_name` is not a valid variable name (got {value})."
            )


class QuestionOptionsDescriptor(BaseDescriptor):
    """Validate that `question_options` is a list, does not exceed the min/max lengths, and has unique items.

    >>> import warnings
    >>> q_class = QuestionOptionsDescriptor.example()
    >>> with warnings.catch_warnings(record=True) as w:
    ...     _ = q_class(["a ", "b", "c"])  # Has trailing space
    ...     assert len(w) == 1
    ...     assert "trailing whitespace" in str(w[0].message)

    >>> _ = q_class(["a", "b", "c", "d", "d"])
    Traceback (most recent call last):
    ...
    edsl.exceptions.questions.QuestionCreationValidationError: Question options must be unique (got ['a', 'b', 'c', 'd', 'd']).

    We allow dynamic question options, which are strings of the form '{{ question_options }}'.

    >>> _ = q_class("{{dynamic_options}}")
    >>> _ = q_class("dynamic_options")
    Traceback (most recent call last):
    ...
    edsl.exceptions.questions.QuestionCreationValidationError: ...
    """

    @classmethod
    def example(cls):
        class TestQuestion:
            question_options = QuestionOptionsDescriptor()

            def __init__(self, question_options: List[str]):
                self.question_options = question_options

        return TestQuestion

    def __init__(
        self,
        num_choices: int = None,
        linear_scale: bool = False,
        q_budget: bool = False,
    ):
        """Initialize the descriptor."""
        self.num_choices = num_choices
        self.linear_scale = linear_scale
        self.q_budget = q_budget

    def validate(self, value: Any, instance) -> None:
        """Validate the question options."""
        if isinstance(value, str):
            # Check if the string is a dynamic question option
            if "{{" in value and "}}" in value:
                return None
            else:
                raise QuestionCreationValidationError(
                    f"Dynamic question options must have jinja2 braces - instead received: {value}."
                )
        if not isinstance(value, list):
            raise QuestionCreationValidationError(
                f"Question options must be a list (got {value})."
            )
        # if len(value) > Settings.MAX_NUM_OPTIONS:
        #     raise QuestionCreationValidationError(
        #         f"Too many question options (got {value})."
        #     )
        if len(value) < Settings.MIN_NUM_OPTIONS:
            raise QuestionCreationValidationError(
                f"Too few question options (got {value})."
            )
        # handle the case when question_options is a list of lists (a list of list can be converted to set)
        tmp_value = [str(x) for x in value]
        if len(tmp_value) != len(set(tmp_value)):
            raise QuestionCreationValidationError(
                f"Question options must be unique (got {value})."
            )
        if not self.linear_scale:
            if not self.q_budget:
                pass
            #     if not (
            #         value
            #         and all(type(x) == type(value[0]) for x in value)
            #         and isinstance(value[0], (str, list, int, float))
            #     ):
            #         raise QuestionCreationValidationError(
            #             f"Question options must be all same type (got {value}).)"
            #         )
            else:
                if not all(isinstance(x, (str)) for x in value):
                    raise QuestionCreationValidationError(
                        f"Question options must be strings (got {value}).)"
                    )
            if not all(
                [
                    type(option) != str
                    or (len(option) >= 1 and len(option) < Settings.MAX_OPTION_LENGTH)
                    for option in value
                ]
            ):
                raise QuestionCreationValidationError(
                    f"All question options must be at least 1 character long but less than {Settings.MAX_OPTION_LENGTH} characters long (got {value})."
                )

            # Check for trailing whitespace in string options
            if any(isinstance(x, str) and (x != x.strip()) for x in value):
                import warnings

                warnings.warn(
                    "Some question options contain trailing whitespace. This may cause unexpected behavior.",
                    UserWarning,
                )

        if hasattr(instance, "min_selections") and instance.min_selections != None:
            if instance.min_selections > len(value):
                raise QuestionCreationValidationError(
                    f"You asked for at least {instance.min_selections} selections, but provided fewer options (got {value})."
                )
        if hasattr(instance, "max_selections") and instance.max_selections != None:
            if instance.max_selections > len(value):
                raise QuestionCreationValidationError(
                    f"You asked for at most {instance.max_selections} selections, but provided fewer options (got {value})."
                )
        if self.num_choices is not None:
            if len(value) != self.num_choices:
                raise QuestionCreationValidationError(
                    f"You asked for {self.num_choices} selections, but provided {len(value)} options."
                )
        if self.linear_scale:
            if sorted(value) != list(range(min(value), max(value) + 1)):
                raise QuestionCreationValidationError(
                    f"LinearScale.question_options must be a list of successive integers, e.g. [1, 2, 3] (got {value})."
                )


class QuestionTextDescriptor(BaseDescriptor):
    """Validate that the `question_text` attribute is a string.


    >>> class TestQuestion:
    ...     question_text = QuestionTextDescriptor()
    ...     def __init__(self, question_text: str):
    ...         self.question_text = question_text

    >>> _ = TestQuestion("What is the capital of France?")
    >>> _ = TestQuestion("What is the capital of France? {{variable}}")
    >>> _ = TestQuestion("What is the capital of France? {{variable name}}")
    Traceback (most recent call last):
    ...
    edsl.exceptions.questions.QuestionCreationValidationError: Question text contains an invalid identifier: 'variable name'
    """

    def validate(self, value, instance):
        """Validate the value is a string."""
        # if len(value) > Settings.MAX_QUESTION_LENGTH:
        #     raise Exception("Question is too long!")
        if len(value) < 1:
            raise Exception("Question is too short!")
        if not isinstance(value, str):
            raise Exception("Question must be a string!")
        
        #value = textwrap.dedent(value).strip()

        if contains_single_braced_substring(value):
            import warnings

            # # warnings.warn(
            # #     f"WARNING: Question text contains a single-braced substring: If you intended to parameterize the question with a Scenario this should be changed to a double-braced substring, e.g. {{variable}}.\nSee details on constructing Scenarios in the docs: https://docs.expectedparrot.com/en/latest/scenarios.html",
            # #     UserWarning,
            # # )
            warnings.warn(
                "WARNING: Question text contains a single-braced substring. "
                "If you intended to parameterize the question with a Scenario, this will "
                "be changed to a double-braced substring, e.g. {{variable}}.\n"
                "See details on constructing Scenarios in the docs: "
                "https://docs.expectedparrot.com/en/latest/scenarios.html",
                UserWarning,
            )
            # Automatically replace single braces with double braces
            # This is here because if the user is using an f-string, the double brace will get converted to a single brace.
            # This undoes that.
            # value = re.sub(r"\{([^\{\}]+)\}", r"{{\1}}", value)
            return value

        # iterate through all doubles braces and check if they are valid python identifiers
        for match in re.finditer(r"\{\{([^\{\}]+)\}\}", value):
            if " " in match.group(1).strip():
                raise QuestionCreationValidationError(
                    f"Question text contains an invalid identifier: '{match.group(1)}'"
                )

        return None


class ValueTypesDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        """Validate the value is a list of strings or None."""
        if value is None:  # Allow None as a valid value
            return None
        if not isinstance(value, list):
            raise QuestionCreationValidationError(
                f"`value_types` must be a list or None (got {value})."
            )
        # Convert all items in the list to strings
        return [str(item) for item in value]


class ValueDescriptionsDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        """Validate the value is a list of strings or None."""
        if value is None:  # Allow None as a valid value
            return None
        if not isinstance(value, list):
            raise QuestionCreationValidationError(
                f"`value_descriptions` must be a list or None (got {value})."
            )
        if not all(isinstance(x, str) for x in value):
            raise QuestionCreationValidationError(
                f"`value_descriptions` must be a list of strings (got {value})."
            )
        return value


class AnswerKeysDescriptor(BaseDescriptor):
    """Validate that the `answer_keys` attribute is a list of strings or integers."""

    def validate(self, value, instance):
        """Validate the value is a list of strings or integers."""
        if not isinstance(value, list):
            raise QuestionCreationValidationError(
                f"`answer_keys` must be a list (got {value})."
            )
        if not all(isinstance(x, (str, int)) for x in value):
            raise QuestionCreationValidationError(
                f"`answer_keys` must be a list of strings or integers (got {value})."
            )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
import copy
import itertools
from typing import Optional, List, Callable, Type, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from edsl.questions.QuestionBase import QuestionBase
    from edsl.scenarios.ScenarioList import ScenarioList


class QuestionBaseGenMixin:
    """Mixin for QuestionBase.
    
    This mostly has functions that are used to generate new questions from existing ones.
    
    """

    def copy(self) -> QuestionBase:
        """Return a deep copy of the question.

        >>> from edsl.questions import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite color?")
        >>> q2 = q.copy()
        >>> q2.question_name
        'color'

        """
        return copy.deepcopy(self)

    def option_permutations(self) -> list[QuestionBase]:
        """Return a list of questions with all possible permutations of the options.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice as Q
        >>> len(Q.example().option_permutations())
        24
        """

        if not hasattr(self, "question_options"):
            return [self]

        questions = []
        for index, permutation in enumerate(
            itertools.permutations(self.question_options)
        ):
            question = copy.deepcopy(self)
            question.question_options = list(permutation)
            question.question_name = f"{self.question_name}_{index}"
            questions.append(question)
        return questions

    def draw(self) -> "QuestionBase":
        """Return a new question with a randomly selected permutation of the options.

        If the question has no options, returns a copy of the original question.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice as Q
        >>> q = Q.example()
        >>> drawn = q.draw()
        >>> len(drawn.question_options) == len(q.question_options)
        True
        >>> q is drawn
        False
        """

        if not hasattr(self, "question_options"):
            return copy.deepcopy(self)

        import random

        question = copy.deepcopy(self)
        question.question_options = list(
            random.sample(self.question_options, len(self.question_options))
        )
        return question

    def loop(self, scenario_list: ScenarioList) -> List[QuestionBase]:
        """Return a list of questions with the question name modified for each scenario.

        :param scenario_list: The list of scenarios to loop through.

        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.scenarios.ScenarioList import ScenarioList
        >>> q = QuestionFreeText(question_text = "What are your thoughts on: {{ subject}}?", question_name = "base_{{subject}}")
        >>> len(q.loop(ScenarioList.from_list("subject", ["Math", "Economics", "Chemistry"])))
        3
        """
        from edsl.questions.loop_processor import LoopProcessor

        lp = LoopProcessor(self)
        return lp.process_templates(scenario_list)

    class MaxTemplateNestingExceeded(Exception):
        """Raised when template rendering exceeds maximum allowed nesting level."""
        pass

    def render(self, replacement_dict: dict, return_dict: bool = False) -> Union["QuestionBase", dict]:
        """Render the question components as jinja2 templates with the replacement dictionary.
        Handles nested template variables by recursively rendering until all variables are resolved.
        
        Raises:
            MaxTemplateNestingExceeded: If template nesting exceeds MAX_NESTING levels
        
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite {{ thing }}?")
        >>> q.render({"thing": "color"})
        Question('free_text', question_name = \"""color\""", question_text = \"""What is your favorite color?\""")

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> q = QuestionMultipleChoice(question_name = "color", question_text = "What is your favorite {{ thing }}?", question_options = ["red", "blue", "green"])
        >>> from edsl.scenarios.Scenario import Scenario
        >>> q.render(Scenario({"thing": "color"})).data
        {'question_name': 'color', 'question_text': 'What is your favorite color?', 'question_options': ['red', 'blue', 'green']}

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> q = QuestionMultipleChoice(question_name = "color", question_text = "What is your favorite {{ thing }}?", question_options = ["red", "blue", "green"])
        >>> q.render({"thing": 1}).data
        {'question_name': 'color', 'question_text': 'What is your favorite 1?', 'question_options': ['red', 'blue', 'green']}


        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> from edsl.scenarios.Scenario import Scenario
        >>> q = QuestionMultipleChoice(question_name = "color", question_text = "What is your favorite {{ thing }}?", question_options = ["red", "blue", "green"])
        >>> q.render(Scenario({"thing": "color of {{ object }}", "object":"water"})).data
        {'question_name': 'color', 'question_text': 'What is your favorite color of water?', 'question_options': ['red', 'blue', 'green']}

        
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "infinite", question_text = "This has {{ a }}")
        >>> q.render({"a": "{{ b }}", "b": "{{ a }}"}) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        edsl.questions.question_base_gen_mixin.QuestionBaseGenMixin.MaxTemplateNestingExceeded:...
        """
        from jinja2 import Environment, meta
        from edsl.scenarios.Scenario import Scenario

        MAX_NESTING = 10  # Maximum allowed nesting levels
        
        strings_only_replacement_dict = {
            k: v for k, v in replacement_dict.items() if not isinstance(v, Scenario)
        }

        def _has_unrendered_variables(template_str: str, env: Environment) -> bool:
            """Check if the template string has any unrendered variables."""
            if not isinstance(template_str, str):
                return False
            ast = env.parse(template_str)
            return bool(meta.find_undeclared_variables(ast))

        def render_string(value: str) -> str:
            if value is None or not isinstance(value, str):
                return value
            
            try:
                env = Environment()
                result = value
                nesting_count = 0
                
                while _has_unrendered_variables(result, env):
                    if nesting_count >= MAX_NESTING:
                        raise self.MaxTemplateNestingExceeded(
                            f"Template rendering exceeded {MAX_NESTING} levels of nesting. "
                            f"Current value: {result}"
                        )
                    
                    template = env.from_string(result)
                    new_result = template.render(strings_only_replacement_dict)
                    if new_result == result:  # Break if no changes made
                        break
                    result = new_result
                    nesting_count += 1
                
                return result
            except self.MaxTemplateNestingExceeded:
                raise
            except Exception as e:
                import warnings
                warnings.warn("Failed to render string: " + value)
                return value
        if return_dict:
            return self._apply_function_dict(render_string)
        else:
            return self.apply_function(render_string)
      
    def apply_function(
        self, func: Callable, exclude_components: Optional[List[str]] = None
    ) -> QuestionBase:
        from edsl.questions.QuestionBase import QuestionBase
        d = self._apply_function_dict(func, exclude_components)
        return QuestionBase.from_dict(d)

    def _apply_function_dict(
        self, func: Callable, exclude_components: Optional[List[str]] = None
    ) -> dict:
        """Apply a function to the question parts, excluding certain components.

        :param func: The function to apply to the question parts.
        :param exclude_components: The components to exclude from the function application.

        >>> from edsl.questions import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite color?")
        >>> shouting = lambda x: x.upper()
        >>> q.apply_function(shouting)
        Question('free_text', question_name = \"""color\""", question_text = \"""WHAT IS YOUR FAVORITE COLOR?\""")

        >>> q.apply_function(shouting, exclude_components = ["question_type"])
        Question('free_text', question_name = \"""COLOR\""", question_text = \"""WHAT IS YOUR FAVORITE COLOR?\""")

        """

        if exclude_components is None:
            exclude_components = ["question_name", "question_type"]

        d = copy.deepcopy(self.to_dict(add_edsl_version=False))
        for key, value in d.items():
            if key in exclude_components:
                continue
            if isinstance(value, dict):
                for k, v in value.items():
                    value[k] = func(v)
                d[key] = value
                continue
            if isinstance(value, list):
                value = [func(v) for v in value]
                d[key] = value
                continue
            d[key] = func(value)
        return d


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""A list of Scenarios to be used in a survey."""

from __future__ import annotations
from typing import (
    Any,
    Optional,
    Union,
    List,
    Callable,
    Literal,
    TYPE_CHECKING,
)

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

import csv
import random
from io import StringIO
import inspect
from collections import UserList, defaultdict
from collections.abc import Iterable

if TYPE_CHECKING:
    from urllib.parse import ParseResult
    from edsl.results.Dataset import Dataset
    from edsl.jobs.Jobs import Jobs
    from edsl.surveys.Survey import Survey
    from edsl.questions.QuestionBase import QuestionBase


from simpleeval import EvalWithCompoundTypes, NameNotDefined  # type: ignore

from tabulate import tabulate_formats

from edsl.Base import Base
from edsl.utilities.remove_edsl_version import remove_edsl_version

from edsl.scenarios.Scenario import Scenario
from edsl.scenarios.ScenarioListPdfMixin import ScenarioListPdfMixin
from edsl.scenarios.ScenarioListExportMixin import ScenarioListExportMixin
from edsl.utilities.naming_utilities import sanitize_string
from edsl.utilities.is_valid_variable_name import is_valid_variable_name
from edsl.exceptions.scenarios import ScenarioError

from edsl.scenarios.directory_scanner import DirectoryScanner


class ScenarioListMixin(ScenarioListPdfMixin, ScenarioListExportMixin):
    pass


if TYPE_CHECKING:
    from edsl.results.Dataset import Dataset

TableFormat: TypeAlias = Literal[
    "plain",
    "simple",
    "github",
    "grid",
    "fancy_grid",
    "pipe",
    "orgtbl",
    "rst",
    "mediawiki",
    "html",
    "latex",
    "latex_raw",
    "latex_booktabs",
    "tsv",
]


class ScenarioList(Base, UserList, ScenarioListMixin):
    """Class for creating a list of scenarios to be used in a survey."""

    __documentation__ = (
        "https://docs.expectedparrot.com/en/latest/scenarios.html#scenariolist"
    )

    def __init__(
        self, data: Optional[list] = None, codebook: Optional[dict[str, str]] = None
    ):
        """Initialize the ScenarioList class."""
        if data is not None:
            super().__init__(data)
        else:
            super().__init__([])
        self.codebook = codebook or {}

    def unique(self) -> ScenarioList:
        """Return a list of unique scenarios.

        >>> s = ScenarioList([Scenario({'a': 1}), Scenario({'a': 1}), Scenario({'a': 2})])
        >>> s.unique()
        ScenarioList([Scenario({'a': 1}), Scenario({'a': 2})])
        """
        return ScenarioList(list(set(self)))

    @property
    def has_jinja_braces(self) -> bool:
        """Check if the ScenarioList has Jinja braces."""
        return any([scenario.has_jinja_braces for scenario in self])

    def _convert_jinja_braces(self) -> ScenarioList:
        """Convert Jinja braces to Python braces."""
        return ScenarioList([scenario._convert_jinja_braces() for scenario in self])

    def give_valid_names(self, existing_codebook: dict = None) -> ScenarioList:
        """Give valid names to the scenario keys, using an existing codebook if provided.

        Args:
            existing_codebook (dict, optional): Existing mapping of original keys to valid names.
                Defaults to None.

        Returns:
            ScenarioList: A new ScenarioList with valid variable names and updated codebook.

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s.give_valid_names()
        ScenarioList([Scenario({'a': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s = ScenarioList([Scenario({'are you there John?': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s.give_valid_names()
        ScenarioList([Scenario({'john': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s.give_valid_names({'are you there John?': 'custom_name'})
        ScenarioList([Scenario({'custom_name': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        """
        codebook = existing_codebook.copy() if existing_codebook else {}
        new_scenarios = []

        for scenario in self:
            new_scenario = {}
            for key in scenario:
                if is_valid_variable_name(key):
                    new_scenario[key] = scenario[key]
                    continue

                if key in codebook:
                    new_key = codebook[key]
                else:
                    new_key = sanitize_string(key)
                    if not is_valid_variable_name(new_key):
                        new_key = f"var_{len(codebook)}"
                    codebook[key] = new_key

                new_scenario[new_key] = scenario[key]

            new_scenarios.append(Scenario(new_scenario))

        return ScenarioList(new_scenarios, codebook)

    def unpivot(
        self,
        id_vars: Optional[List[str]] = None,
        value_vars: Optional[List[str]] = None,
    ) -> ScenarioList:
        """
        Unpivot the ScenarioList, allowing for id variables to be specified.

        Parameters:
        id_vars (list): Fields to use as identifier variables (kept in each entry)
        value_vars (list): Fields to unpivot. If None, all fields not in id_vars will be used.

        Example:
        >>> s = ScenarioList([
        ...     Scenario({'id': 1, 'year': 2020, 'a': 10, 'b': 20}),
        ...     Scenario({'id': 2, 'year': 2021, 'a': 15, 'b': 25})
        ... ])
        >>> s.unpivot(id_vars=['id', 'year'], value_vars=['a', 'b'])
        ScenarioList([Scenario({'id': 1, 'year': 2020, 'variable': 'a', 'value': 10}), Scenario({'id': 1, 'year': 2020, 'variable': 'b', 'value': 20}), Scenario({'id': 2, 'year': 2021, 'variable': 'a', 'value': 15}), Scenario({'id': 2, 'year': 2021, 'variable': 'b', 'value': 25})])
        """
        if id_vars is None:
            id_vars = []
        if value_vars is None:
            value_vars = [field for field in self[0].keys() if field not in id_vars]

        new_scenarios = []
        for scenario in self:
            for var in value_vars:
                new_scenario = {id_var: scenario[id_var] for id_var in id_vars}
                new_scenario["variable"] = var
                new_scenario["value"] = scenario[var]
                new_scenarios.append(Scenario(new_scenario))

        return ScenarioList(new_scenarios)

    def sem_filter(self, language_predicate: str) -> ScenarioList:
        """Filter the ScenarioList based on a language predicate.

        :param language_predicate: The language predicate to use.

        Inspired by:
        @misc{patel2024semanticoperators,
            title={Semantic Operators: A Declarative Model for Rich, AI-based Analytics Over Text Data},
            author={Liana Patel and Siddharth Jha and Parth Asawa and Melissa Pan and Carlos Guestrin and Matei Zaharia},
            year={2024},
            eprint={2407.11418},
            archivePrefix={arXiv},
            primaryClass={cs.DB},
            url={https://arxiv.org/abs/2407.11418},
            }
        """
        from edsl import QuestionYesNo

        new_scenario_list = self.duplicate()
        q = QuestionYesNo(
            question_text=language_predicate, question_name="binary_outcome"
        )
        results = q.by(new_scenario_list).run(verbose=False)
        new_scenario_list = new_scenario_list.add_list(
            "criteria", results.select("binary_outcome").to_list()
        )
        return new_scenario_list.filter("criteria == 'Yes'").drop("criteria")

    def pivot(
        self,
        id_vars: List[str] = None,
        var_name="variable",
        value_name="value",
    ) -> ScenarioList:
        """
        Pivot the ScenarioList from long to wide format.

        Parameters:
        id_vars (list): Fields to use as identifier variables
        var_name (str): Name of the variable column (default: 'variable')
        value_name (str): Name of the value column (default: 'value')

        Example:
        >>> s = ScenarioList([
        ...     Scenario({'id': 1, 'year': 2020, 'variable': 'a', 'value': 10}),
        ...     Scenario({'id': 1, 'year': 2020, 'variable': 'b', 'value': 20}),
        ...     Scenario({'id': 2, 'year': 2021, 'variable': 'a', 'value': 15}),
        ...     Scenario({'id': 2, 'year': 2021, 'variable': 'b', 'value': 25})
        ... ])
        >>> s.pivot(id_vars=['id', 'year'])
        ScenarioList([Scenario({'id': 1, 'year': 2020, 'a': 10, 'b': 20}), Scenario({'id': 2, 'year': 2021, 'a': 15, 'b': 25})])
        """
        pivoted_dict = {}

        for scenario in self:
            # Create a tuple of id values to use as a key
            id_key = tuple(scenario[id_var] for id_var in id_vars)

            # If this combination of id values hasn't been seen before, initialize it
            if id_key not in pivoted_dict:
                pivoted_dict[id_key] = {id_var: scenario[id_var] for id_var in id_vars}

            # Add the variable-value pair to the dict
            variable = scenario[var_name]
            value = scenario[value_name]
            pivoted_dict[id_key][variable] = value

        # Convert the dict of dicts to a list of Scenarios
        pivoted_scenarios = [
            Scenario(dict(zip(id_vars, id_key), **values))
            for id_key, values in pivoted_dict.items()
        ]

        return ScenarioList(pivoted_scenarios)

    def group_by(
        self, id_vars: List[str], variables: List[str], func: Callable
    ) -> ScenarioList:
        """
        Group the ScenarioList by id_vars and apply a function to the specified variables.

        :param id_vars: Fields to use as identifier variables
        :param variables: Fields to group and aggregate
        :param func: Function to apply to the grouped variables

        Returns:
        ScenarioList: A new ScenarioList with the grouped and aggregated results

        Example:
        >>> def avg_sum(a, b):
        ...     return {'avg_a': sum(a) / len(a), 'sum_b': sum(b)}
        >>> s = ScenarioList([
        ...     Scenario({'group': 'A', 'year': 2020, 'a': 10, 'b': 20}),
        ...     Scenario({'group': 'A', 'year': 2021, 'a': 15, 'b': 25}),
        ...     Scenario({'group': 'B', 'year': 2020, 'a': 12, 'b': 22}),
        ...     Scenario({'group': 'B', 'year': 2021, 'a': 17, 'b': 27})
        ... ])
        >>> s.group_by(id_vars=['group'], variables=['a', 'b'], func=avg_sum)
        ScenarioList([Scenario({'group': 'A', 'avg_a': 12.5, 'sum_b': 45}), Scenario({'group': 'B', 'avg_a': 14.5, 'sum_b': 49})])
        """
        # Check if the function is compatible with the specified variables
        func_params = inspect.signature(func).parameters
        if len(func_params) != len(variables):
            raise ScenarioError(
                f"Function {func.__name__} expects {len(func_params)} arguments, but {len(variables)} variables were provided"
            )

        # Group the scenarios
        grouped: dict[str, list] = defaultdict(lambda: defaultdict(list))
        for scenario in self:
            key = tuple(scenario[id_var] for id_var in id_vars)
            for var in variables:
                grouped[key][var].append(scenario[var])

        # Apply the function to each group
        result = []
        for key, group in grouped.items():
            try:
                aggregated = func(*[group[var] for var in variables])
            except Exception as e:
                raise ScenarioError(f"Error applying function to group {key}: {str(e)}")

            if not isinstance(aggregated, dict):
                raise ScenarioError(
                    f"Function {func.__name__} must return a dictionary"
                )

            new_scenario = dict(zip(id_vars, key))
            new_scenario.update(aggregated)
            result.append(Scenario(new_scenario))

        return ScenarioList(result)

    @property
    def parameters(self) -> set:
        """Return the set of parameters in the ScenarioList

        Example:

        >>> s = ScenarioList([Scenario({'a': 1}), Scenario({'b': 2})])
        >>> s.parameters == {'a', 'b'}
        True
        """
        if len(self) == 0:
            return set()

        return set.union(*[set(s.keys()) for s in self])

    def __hash__(self) -> int:
        """Return the hash of the ScenarioList.

        >>> s = ScenarioList.example()
        >>> hash(s)
        1262252885757976162
        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(sort=True, add_edsl_version=False))

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def __repr__(self):
        return f"ScenarioList({self.data})"

    def __mul__(self, other: ScenarioList) -> ScenarioList:
        """Takes the cross product of two ScenarioLists.

        >>> s1 = ScenarioList.from_list("a", [1, 2])
        >>> s2 = ScenarioList.from_list("b", [3, 4])
        >>> s1 * s2
        ScenarioList([Scenario({'a': 1, 'b': 3}), Scenario({'a': 1, 'b': 4}), Scenario({'a': 2, 'b': 3}), Scenario({'a': 2, 'b': 4})])
        """
        from itertools import product
        from edsl import Scenario
        if isinstance(other, Scenario):
            other = ScenarioList([other])
        elif not isinstance(other, ScenarioList):
            raise TypeError(f"Cannot multiply ScenarioList with {type(other)}")

        new_sl = []
        for s1, s2 in list(product(self, other)):
            new_sl.append(s1 + s2)
        return ScenarioList(new_sl)

    def times(self, other: ScenarioList) -> ScenarioList:
        """Takes the cross product of two ScenarioLists.

        Example:

        >>> s1 = ScenarioList([Scenario({'a': 1}), Scenario({'a': 2})])
        >>> s2 = ScenarioList([Scenario({'b': 1}), Scenario({'b': 2})])
        >>> s1.times(s2)
        ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2}), Scenario({'a': 2, 'b': 1}), Scenario({'a': 2, 'b': 2})])
        """
        return self.__mul__(other)

    def shuffle(self, seed: Optional[str] = None) -> ScenarioList:
        """Shuffle the ScenarioList.

        >>> s = ScenarioList.from_list("a", [1,2,3,4])
        >>> s.shuffle(seed = "1234")
        ScenarioList([Scenario({'a': 1}), Scenario({'a': 4}), Scenario({'a': 3}), Scenario({'a': 2})])
        """
        sl = self.duplicate()
        if seed:
            random.seed(seed)
        random.shuffle(sl.data)
        return sl

    def sample(self, n: int, seed: Optional[str] = None) -> ScenarioList:
        """Return a random sample from the ScenarioList

        >>> s = ScenarioList.from_list("a", [1,2,3,4,5,6])
        >>> s.sample(3, seed = "edsl")
        ScenarioList([Scenario({'a': 2}), Scenario({'a': 1}), Scenario({'a': 3})])
        """
        if seed:
            random.seed(seed)

        sl = self.duplicate()
        return ScenarioList(random.sample(sl.data, n))

    def expand(self, expand_field: str, number_field: bool = False) -> ScenarioList:
        """Expand the ScenarioList by a field.

        :param expand_field: The field to expand.
        :param number_field: Whether to add a field with the index of the value

        Example:

        >>> s = ScenarioList( [ Scenario({'a':1, 'b':[1,2]}) ] )
        >>> s.expand('b')
        ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2})])
        >>> s.expand('b', number_field=True)
        ScenarioList([Scenario({'a': 1, 'b': 1, 'b_number': 1}), Scenario({'a': 1, 'b': 2, 'b_number': 2})])
        """
        new_scenarios = []
        for scenario in self:
            values = scenario[expand_field]
            if not isinstance(values, Iterable) or isinstance(values, str):
                values = [values]
            for index, value in enumerate(values):
                new_scenario = scenario.copy()
                new_scenario[expand_field] = value
                if number_field:
                    new_scenario[expand_field + "_number"] = index + 1
                new_scenarios.append(new_scenario)
        return ScenarioList(new_scenarios)

    def concatenate(self, fields: List[str], separator: str = ";") -> ScenarioList:
        """Concatenate specified fields into a single field.

        :param fields: The fields to concatenate.
        :param separator: The separator to use.

        Returns:
            ScenarioList: A new ScenarioList with concatenated fields.

        Example:
            >>> s = ScenarioList([Scenario({'a': 1, 'b': 2, 'c': 3}), Scenario({'a': 4, 'b': 5, 'c': 6})])
            >>> s.concatenate(['a', 'b', 'c'])
            ScenarioList([Scenario({'concat_a_b_c': '1;2;3'}), Scenario({'concat_a_b_c': '4;5;6'})])
        """
        new_scenarios = []
        for scenario in self:
            new_scenario = scenario.copy()
            concat_values = []
            for field in fields:
                if field in new_scenario:
                    concat_values.append(str(new_scenario[field]))
                    del new_scenario[field]

            new_field_name = f"concat_{'_'.join(fields)}"
            new_scenario[new_field_name] = separator.join(concat_values)
            new_scenarios.append(new_scenario)

        return ScenarioList(new_scenarios)

    def unpack_dict(
        self, field: str, prefix: Optional[str] = None, drop_field: bool = False
    ) -> ScenarioList:
        """Unpack a dictionary field into separate fields.

        :param field: The field to unpack.
        :param prefix: An optional prefix to add to the new fields.
        :param drop_field: Whether to drop the original field.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': {'c': 2, 'd': 3}})])
        >>> s.unpack_dict('b')
        ScenarioList([Scenario({'a': 1, 'b': {'c': 2, 'd': 3}, 'c': 2, 'd': 3})])
        >>> s.unpack_dict('b', prefix='new_')
        ScenarioList([Scenario({'a': 1, 'b': {'c': 2, 'd': 3}, 'new_c': 2, 'new_d': 3})])
        """
        new_scenarios = []
        for scenario in self:
            new_scenario = scenario.copy()
            for key, value in scenario[field].items():
                if prefix:
                    new_scenario[prefix + key] = value
                else:
                    new_scenario[key] = value
            if drop_field:
                new_scenario.pop(field)
            new_scenarios.append(new_scenario)
        return ScenarioList(new_scenarios)

    def transform(
        self, field: str, func: Callable, new_name: Optional[str] = None
    ) -> ScenarioList:
        """Transform a field using a function.

        :param field: The field to transform.
        :param func: The function to apply to the field.
        :param new_name: An optional new name for the transformed field.

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s.transform('b', lambda x: x + 1)
        ScenarioList([Scenario({'a': 1, 'b': 3}), Scenario({'a': 1, 'b': 2})])

        """
        new_scenarios = []
        for scenario in self:
            new_scenario = scenario.copy()
            new_scenario[new_name or field] = func(scenario[field])
            new_scenarios.append(new_scenario)
        return ScenarioList(new_scenarios)

    def mutate(
        self, new_var_string: str, functions_dict: Optional[dict[str, Callable]] = None
    ) -> ScenarioList:
        """
        Return a new ScenarioList with a new variable added.

        :param new_var_string: A string with the new variable assignment.
        :param functions_dict: A dictionary of functions to use in the assignment.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s.mutate("c = a + b")
        ScenarioList([Scenario({'a': 1, 'b': 2, 'c': 3}), Scenario({'a': 1, 'b': 1, 'c': 2})])

        """
        if "=" not in new_var_string:
            raise ScenarioError(
                f"Mutate requires an '=' in the string, but '{new_var_string}' doesn't have one."
            )
        raw_var_name, expression = new_var_string.split("=", 1)
        var_name = raw_var_name.strip()
        from edsl.utilities.utilities import is_valid_variable_name

        if not is_valid_variable_name(var_name):
            raise ScenarioError(f"{var_name} is not a valid variable name.")

        # create the evaluator
        functions_dict = functions_dict or {}

        def create_evaluator(scenario) -> EvalWithCompoundTypes:
            return EvalWithCompoundTypes(names=scenario, functions=functions_dict)

        def new_scenario(old_scenario: Scenario, var_name: str) -> Scenario:
            evaluator = create_evaluator(old_scenario)
            value = evaluator.eval(expression)
            new_s = old_scenario.copy()
            new_s[var_name] = value
            return new_s

        try:
            new_data = [new_scenario(s, var_name) for s in self]
        except Exception as e:
            raise ScenarioError(f"Error in mutate. Exception:{e}")

        return ScenarioList(new_data)

    def order_by(self, *fields: str, reverse: bool = False) -> ScenarioList:
        """Order the scenarios by one or more fields.

        :param fields: The fields to order by.
        :param reverse: Whether to reverse the order.
        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 2}), Scenario({'a': 1, 'b': 1})])
        >>> s.order_by('b', 'a')
        ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2})])
        """

        def get_sort_key(scenario: Any) -> tuple:
            return tuple(scenario[field] for field in fields)

        return ScenarioList(sorted(self, key=get_sort_key, reverse=reverse))

    def duplicate(self) -> ScenarioList:
        """Return a copy of the ScenarioList.

        >>> sl = ScenarioList.example()
        >>> sl_copy = sl.duplicate()
        >>> sl == sl_copy
        True
        >>> sl is sl_copy
        False
        """
        return ScenarioList([scenario.copy() for scenario in self])

    def filter(self, expression: str) -> ScenarioList:
        """
        Filter a list of scenarios based on an expression.

        :param expression: The expression to filter by.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2})])
        >>> s.filter("b == 2")
        ScenarioList([Scenario({'a': 1, 'b': 2})])
        """
        sl = self.duplicate()
        base_keys = set(self[0].keys())
        keys = set()
        for scenario in sl:
            keys.update(scenario.keys())
        if keys != base_keys:
            import warnings

            warnings.warn(
                "Ragged ScenarioList detected (different keys for different scenario entries). This may cause unexpected behavior."
            )

        def create_evaluator(scenario: Scenario):
            """Create an evaluator for the given result.
            The 'combined_dict' is a mapping of all values for that Result object.
            """
            return EvalWithCompoundTypes(names=scenario)

        try:
            # iterates through all the results and evaluates the expression
            new_data = []
            for scenario in sl:
                if create_evaluator(scenario).eval(expression):
                    new_data.append(scenario)
        except NameNotDefined as e:
            available_fields = ", ".join(self.data[0].keys() if self.data else [])
            raise ScenarioError(
                f"Error in filter: '{e}'\n"
                f"The expression '{expression}' refers to a field that does not exist.\n"
                f"Scenario: {scenario}\n"
                f"Available fields: {available_fields}\n"
                "Check your filter expression or consult the documentation: "
                "https://docs.expectedparrot.com/en/latest/scenarios.html#module-edsl.scenarios.Scenario"
            ) from None
        except Exception as e:
            raise ScenarioError(f"Error in filter. Exception:{e}")

        return ScenarioList(new_data)

    def from_urls(
        self, urls: list[str], field_name: Optional[str] = "text"
    ) -> ScenarioList:
        """Create a ScenarioList from a list of URLs.

        :param urls: A list of URLs.
        :param field_name: The name of the field to store the text from the URLs.

        """
        return ScenarioList([Scenario.from_url(url, field_name) for url in urls])

    def select(self, *fields: str) -> ScenarioList:
        """
        Selects scenarios with only the references fields.

        :param fields: The fields to select.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2})])
        >>> s.select('a')
        ScenarioList([Scenario({'a': 1}), Scenario({'a': 1})])
        """
        from edsl.scenarios.scenario_selector import ScenarioSelector

        return ScenarioSelector(self).select(*fields)

    def drop(self, *fields: str) -> ScenarioList:
        """Drop fields from the scenarios.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2})])
        >>> s.drop('a')
        ScenarioList([Scenario({'b': 1}), Scenario({'b': 2})])
        """
        sl = self.duplicate()
        return ScenarioList([scenario.drop(fields) for scenario in sl])

    def keep(self, *fields: str) -> ScenarioList:
        """Keep only the specified fields in the scenarios.

        :param fields: The fields to keep.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 1}), Scenario({'a': 1, 'b': 2})])
        >>> s.keep('a')
        ScenarioList([Scenario({'a': 1}), Scenario({'a': 1})])
        """
        sl = self.duplicate()
        return ScenarioList([scenario.keep(fields) for scenario in sl])

    @classmethod
    def from_list(
        cls, name: str, values: list, func: Optional[Callable] = None
    ) -> ScenarioList:
        """Create a ScenarioList from a list of values.

        :param name: The name of the field.
        :param values: The list of values.
        :param func: An optional function to apply to the values.

        Example:

        >>> ScenarioList.from_list('name', ['Alice', 'Bob'])
        ScenarioList([Scenario({'name': 'Alice'}), Scenario({'name': 'Bob'})])
        """
        if not func:
            func = lambda x: x
        return cls([Scenario({name: func(value)}) for value in values])

    def table(
        self,
        *fields: str,
        tablefmt: Optional[TableFormat] = None,
        pretty_labels: Optional[dict[str, str]] = None,
    ) -> str:
        """Return the ScenarioList as a table."""

        from tabulate import tabulate_formats

        if tablefmt is not None and tablefmt not in tabulate_formats:
            raise ValueError(
                f"Invalid table format: {tablefmt}",
                f"Valid formats are: {tabulate_formats}",
            )
        return self.to_dataset().table(
            *fields, tablefmt=tablefmt, pretty_labels=pretty_labels
        )

    def tree(self, node_list: Optional[List[str]] = None) -> str:
        """Return the ScenarioList as a tree.

        :param node_list: The list of nodes to include in the tree.
        """
        return self.to_dataset().tree(node_list)

    def _summary(self) -> dict:
        """Return a summary of the ScenarioList.

        >>> ScenarioList.example()._summary()
        {'scenarios': 2, 'keys': ['persona']}
        """
        d = {
            "scenarios": len(self),
            "keys": list(self.parameters),
        }
        return d

    def reorder_keys(self, new_order: List[str]) -> ScenarioList:
        """Reorder the keys in the scenarios.

        :param new_order: The new order of the keys.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': 2}), Scenario({'a': 3, 'b': 4})])
        >>> s.reorder_keys(['b', 'a'])
        ScenarioList([Scenario({'b': 2, 'a': 1}), Scenario({'b': 4, 'a': 3})])
        >>> s.reorder_keys(['a', 'b', 'c'])
        Traceback (most recent call last):
        ...
        AssertionError
        """
        assert set(new_order) == set(self.parameters)

        new_scenarios = []
        for scenario in self:
            new_scenario = Scenario({key: scenario[key] for key in new_order})
            new_scenarios.append(new_scenario)
        return ScenarioList(new_scenarios)

    def to_dataset(self) -> "Dataset":
        """
        Convert the ScenarioList to a Dataset.

        >>> s = ScenarioList.from_list("a", [1,2,3])
        >>> s.to_dataset()
        Dataset([{'a': [1, 2, 3]}])
        >>> s = ScenarioList.from_list("a", [1,2,3]).add_list("b", [4,5,6])
        >>> s.to_dataset()
        Dataset([{'a': [1, 2, 3]}, {'b': [4, 5, 6]}])
        """
        from edsl.results.Dataset import Dataset

        keys = list(self[0].keys())
        for scenario in self:
            new_keys = list(scenario.keys())
            if new_keys != keys:
                keys = list(set(keys + new_keys))
        data = [
            {key: [scenario.get(key, None) for scenario in self.data]} for key in keys
        ]
        return Dataset(data)

    def unpack(
        self, field: str, new_names: Optional[List[str]] = None, keep_original=True
    ) -> ScenarioList:
        """Unpack a field into multiple fields.

        Example:

        >>> s = ScenarioList([Scenario({'a': 1, 'b': [2, True]}), Scenario({'a': 3, 'b': [3, False]})])
        >>> s.unpack('b')
        ScenarioList([Scenario({'a': 1, 'b': [2, True], 'b_0': 2, 'b_1': True}), Scenario({'a': 3, 'b': [3, False], 'b_0': 3, 'b_1': False})])
        >>> s.unpack('b', new_names=['c', 'd'], keep_original=False)
        ScenarioList([Scenario({'a': 1, 'c': 2, 'd': True}), Scenario({'a': 3, 'c': 3, 'd': False})])

        """
        new_names = new_names or [f"{field}_{i}" for i in range(len(self[0][field]))]
        new_scenarios = []
        for scenario in self:
            new_scenario = scenario.copy()
            if len(new_names) == 1:
                new_scenario[new_names[0]] = scenario[field]
            else:
                for i, new_name in enumerate(new_names):
                    new_scenario[new_name] = scenario[field][i]

            if not keep_original:
                del new_scenario[field]
            new_scenarios.append(new_scenario)
        return ScenarioList(new_scenarios)

    @classmethod
    def from_list_of_tuples(self, *names: str, values: List[Tuple]) -> ScenarioList:
        sl = ScenarioList.from_list(names[0], [value[0] for value in values])
        for index, name in enumerate(names[1:]):
            sl = sl.add_list(name, [value[index + 1] for value in values])
        return sl

    def add_list(self, name: str, values: List[Any]) -> ScenarioList:
        """Add a list of values to a ScenarioList.

        Example:

        >>> s = ScenarioList([Scenario({'name': 'Alice'}), Scenario({'name': 'Bob'})])
        >>> s.add_list('age', [30, 25])
        ScenarioList([Scenario({'name': 'Alice', 'age': 30}), Scenario({'name': 'Bob', 'age': 25})])
        """
        sl = self.duplicate()
        if len(values) != len(sl):
            raise ScenarioError(
                f"Length of values ({len(values)}) does not match length of ScenarioList ({len(sl)})"
            )
        for i, value in enumerate(values):
            sl[i][name] = value
        return sl

    @classmethod
    def create_empty_scenario_list(cls, n: int) -> ScenarioList:
        """Create an empty ScenarioList with n scenarios.

        Example:

        >>> ScenarioList.create_empty_scenario_list(3)
        ScenarioList([Scenario({}), Scenario({}), Scenario({})])
        """
        return ScenarioList([Scenario({}) for _ in range(n)])

    def add_value(self, name: str, value: Any) -> ScenarioList:
        """Add a value to all scenarios in a ScenarioList.

        Example:

        >>> s = ScenarioList([Scenario({'name': 'Alice'}), Scenario({'name': 'Bob'})])
        >>> s.add_value('age', 30)
        ScenarioList([Scenario({'name': 'Alice', 'age': 30}), Scenario({'name': 'Bob', 'age': 30})])
        """
        sl = self.duplicate()
        for scenario in sl:
            scenario[name] = value
        return sl
    
    def rename(self, replacement_dict: dict) -> ScenarioList:
        """Rename the fields in the scenarios.

        :param replacement_dict: A dictionary with the old names as keys and the new names as values.

        Example:

        >>> s = ScenarioList([Scenario({'name': 'Alice', 'age': 30}), Scenario({'name': 'Bob', 'age': 25})])
        >>> s.rename({'name': 'first_name', 'age': 'years'})
        ScenarioList([Scenario({'first_name': 'Alice', 'years': 30}), Scenario({'first_name': 'Bob', 'years': 25})])

        """
        new_list = ScenarioList([])
        for obj in self:
            new_obj = obj.rename(replacement_dict)
            new_list.append(new_obj)
        return new_list

    def replace_names(self, new_names: list) -> ScenarioList:
        """Replace the field names in the scenarios with a new list of names.

        :param new_names: A list of new field names to use.

        Example:

        >>> s = ScenarioList([Scenario({'name': 'Alice', 'age': 30}), Scenario({'name': 'Bob', 'age': 25})])
        >>> s.replace_names(['first_name', 'years'])
        ScenarioList([Scenario({'first_name': 'Alice', 'years': 30}), Scenario({'first_name': 'Bob', 'years': 25})])
        """
        if not self:
            return ScenarioList([])
            
        if len(new_names) != len(self[0].keys()):
            raise ScenarioError(
                f"Length of new names ({len(new_names)}) does not match number of fields ({len(self[0].keys())})"
            )

        old_names = list(self[0].keys())
        replacement_dict = dict(zip(old_names, new_names))
        return self.rename(replacement_dict)

    ## NEEDS TO BE FIXED
    # def new_column_names(self, new_names: List[str]) -> ScenarioList:
    #     """Rename the fields in the scenarios.

    #     Example:

    #     >>> s = ScenarioList([Scenario({'name': 'Alice', 'age': 30}), Scenario({'name': 'Bob', 'age': 25})])
    #     >>> s.new_column_names(['first_name', 'years'])
    #     ScenarioList([Scenario({'first_name': 'Alice', 'years': 30}), Scenario({'first_name': 'Bob', 'years': 25})])

    #     """
    #     new_list = ScenarioList([])
    #     for obj in self:
    #         new_obj = obj.new_column_names(new_names)
    #         new_list.append(new_obj)
    #     return new_list

    @classmethod
    def from_sqlite(cls, filepath: str, table: str):
        """Create a ScenarioList from a SQLite database."""
        import sqlite3

        with sqlite3.connect(filepath) as conn:
            cursor = conn.cursor()
            cursor.execute(f"SELECT * FROM {table}")
            columns = [description[0] for description in cursor.description]
            data = cursor.fetchall()
        return cls([Scenario(dict(zip(columns, row))) for row in data])

    @classmethod
    def from_latex(cls, tex_file_path: str):
        with open(tex_file_path, "r") as file:
            lines = file.readlines()

        processed_lines = []
        non_blank_lines = [
            (i, line.strip()) for i, line in enumerate(lines) if line.strip()
        ]

        for index, (line_no, text) in enumerate(non_blank_lines):
            entry = {
                "line_no": line_no + 1,  # Using 1-based index for line numbers
                "text": text,
                "num_words": len(text.split()),
                "num_chars": len(text),
                "line_before": non_blank_lines[index - 1][1] if index > 0 else None,
                "line_after": (
                    non_blank_lines[index + 1][1]
                    if index < len(non_blank_lines) - 1
                    else None
                ),
            }
            processed_lines.append(entry)

        return ScenarioList([Scenario(entry) for entry in processed_lines])

    @classmethod
    def from_google_doc(cls, url: str) -> ScenarioList:
        """Create a ScenarioList from a Google Doc.

        This method downloads the Google Doc as a Word file (.docx), saves it to a temporary file,
        and then reads it using the from_docx class method.

        Args:
            url (str): The URL to the Google Doc.

        Returns:
            ScenarioList: An instance of the ScenarioList class.

        """
        import tempfile
        import requests
        from docx import Document

        if "/edit" in url:
            doc_id = url.split("/d/")[1].split("/edit")[0]
        else:
            raise ValueError("Invalid Google Doc URL format.")

        export_url = f"https://docs.google.com/document/d/{doc_id}/export?format=docx"

        # Download the Google Doc as a Word file (.docx)
        response = requests.get(export_url)
        response.raise_for_status()  # Ensure the request was successful

        # Save the Word file to a temporary file
        with tempfile.NamedTemporaryFile(suffix=".docx", delete=False) as temp_file:
            temp_file.write(response.content)
            temp_filename = temp_file.name

        # Call the from_docx class method with the temporary file
        return cls.from_docx(temp_filename)

    @classmethod
    def from_pandas(cls, df) -> ScenarioList:
        """Create a ScenarioList from a pandas DataFrame.

        Example:

        >>> import pandas as pd
        >>> df = pd.DataFrame({'name': ['Alice', 'Bob'], 'age': [30, 25], 'location': ['New York', 'Los Angeles']})
        >>> ScenarioList.from_pandas(df)
        ScenarioList([Scenario({'name': 'Alice', 'age': 30, 'location': 'New York'}), Scenario({'name': 'Bob', 'age': 25, 'location': 'Los Angeles'})])
        """
        return cls([Scenario(row) for row in df.to_dict(orient="records")])
    

    @classmethod
    def from_dta(cls, filepath: str, include_metadata: bool = True) -> ScenarioList:
        """Create a ScenarioList from a Stata file.
        
        Args:
            filepath (str): Path to the Stata (.dta) file
            include_metadata (bool): If True, extract and preserve variable labels and value labels
                                    as additional metadata in the ScenarioList
        
        Returns:
            ScenarioList: A ScenarioList containing the data from the Stata file
        """
        import pandas as pd
        
        # Read the Stata file with pandas
        df = pd.read_stata(filepath)
        
        # Create the basic ScenarioList
        scenario_list = cls.from_pandas(df)
        
        # Extract and preserve metadata if requested
        if include_metadata:
            # Get variable labels (if any)
            variable_labels = {}
            if hasattr(df, 'variable_labels') and df.variable_labels:
                variable_labels = df.variable_labels
            
            # Get value labels (if any)
            value_labels = {}
            if hasattr(df, 'value_labels') and df.value_labels:
                value_labels = df.value_labels
            
            # Store the metadata in the ScenarioList's codebook
            if variable_labels or value_labels:
                scenario_list.codebook = {
                    'variable_labels': variable_labels,
                    'value_labels': value_labels
                }
        
        return scenario_list

    def from_wikipedia(cls, url: str, table_index: int = 0):
        """
        Extracts a table from a Wikipedia page.

        Parameters:
            url (str): The URL of the Wikipedia page.
            table_index (int): The index of the table to extract (default is 0).

        Returns:
            pd.DataFrame: A DataFrame containing the extracted table.
        # # Example usage
        # url = "https://en.wikipedia.org/wiki/List_of_countries_by_GDP_(nominal)"
        # df = from_wikipedia(url, 0)

        # if not df.empty:
        #     print(df.head())
        # else:
        #     print("Failed to extract table.")


        """
        import pandas as pd
        import requests
        from requests.exceptions import RequestException

        try:
            # Check if the URL is reachable
            response = requests.get(url)
            response.raise_for_status()  # Raises HTTPError for bad responses

            # Extract tables from the Wikipedia page
            tables = pd.read_html(url)

            # Ensure the requested table index is within the range of available tables
            if table_index >= len(tables) or table_index < 0:
                raise IndexError(
                    f"Table index {table_index} is out of range. This page has {len(tables)} table(s)."
                )

            # Return the requested table as a DataFrame
            # return tables[table_index]
            return cls.from_pandas(tables[table_index])

        except RequestException as e:
            print(f"Error fetching the URL: {e}")
        except ValueError as e:
            print(f"Error parsing tables: {e}")
        except IndexError as e:
            print(e)
        except Exception as e:
            print(f"An unexpected error occurred: {e}")

        # Return an empty DataFrame in case of an error
        # return cls.from_pandas(pd.DataFrame())

    def to_key_value(self, field: str, value=None) -> Union[dict, set]:
        """Return the set of values in the field.

        :param field: The field to extract values from.
        :param value: An optional field to use as the value in the key-value pair.

        Example:

        >>> s = ScenarioList([Scenario({'name': 'Alice'}), Scenario({'name': 'Bob'})])
        >>> s.to_key_value('name') == {'Alice', 'Bob'}
        True
        """
        if value is None:
            return {scenario[field] for scenario in self}
        else:
            return {scenario[field]: scenario[value] for scenario in self}

    @classmethod
    def from_excel(
        cls, filename: str, sheet_name: Optional[str] = None
    ) -> ScenarioList:
        """Create a ScenarioList from an Excel file.

        If the Excel file contains multiple sheets and no sheet_name is provided,
        the method will print the available sheets and require the user to specify one.

        Example:

        >>> import tempfile
        >>> import os
        >>> import pandas as pd
        >>> with tempfile.NamedTemporaryFile(delete=False, suffix='.xlsx') as f:
        ...     df1 = pd.DataFrame({
        ...         'name': ['Alice', 'Bob'],
        ...         'age': [30, 25],
        ...         'location': ['New York', 'Los Angeles']
        ...     })
        ...     df2 = pd.DataFrame({
        ...         'name': ['Charlie', 'David'],
        ...         'age': [35, 40],
        ...         'location': ['Chicago', 'Boston']
        ...     })
        ...     with pd.ExcelWriter(f.name) as writer:
        ...         df1.to_excel(writer, sheet_name='Sheet1', index=False)
        ...         df2.to_excel(writer, sheet_name='Sheet2', index=False)
        ...     temp_filename = f.name
        >>> scenario_list = ScenarioList.from_excel(temp_filename, sheet_name='Sheet1')
        >>> len(scenario_list)
        2
        >>> scenario_list[0]['name']
        'Alice'
        >>> scenario_list = ScenarioList.from_excel(temp_filename)  # Should raise an error and list sheets
        Traceback (most recent call last):
        ...
        ValueError: Please provide a sheet name to load data from.
        """
        from edsl.scenarios.Scenario import Scenario
        import pandas as pd

        # Get all sheets
        all_sheets = pd.read_excel(filename, sheet_name=None)

        # If no sheet_name is provided and there is more than one sheet, print available sheets
        if sheet_name is None:
            if len(all_sheets) > 1:
                print("The Excel file contains multiple sheets:")
                for name in all_sheets.keys():
                    print(f"- {name}")
                raise ValueError("Please provide a sheet name to load data from.")
            else:
                # If there is only one sheet, use it
                sheet_name = list(all_sheets.keys())[0]

        # Load the specified or determined sheet
        df = pd.read_excel(filename, sheet_name=sheet_name)

        observations = []
        for _, row in df.iterrows():
            observations.append(Scenario(row.to_dict()))

        return cls(observations)

    @classmethod
    def from_google_sheet(cls, url: str, sheet_name: str = None, column_names: Optional[List[str]]= None) -> ScenarioList:
        """Create a ScenarioList from a Google Sheet.

        This method downloads the Google Sheet as an Excel file, saves it to a temporary file,
        and then reads it using the from_excel class method.

        Args:
            url (str): The URL to the Google Sheet.
            sheet_name (str, optional): The name of the sheet to load. If None, the method will behave
                                        the same as from_excel regarding multiple sheets.
            column_names (List[str], optional): If provided, use these names for the columns instead
                                              of the default column names from the sheet.

        Returns:
            ScenarioList: An instance of the ScenarioList class.

        """
        import pandas as pd
        import tempfile
        import requests

        if "/edit" in url:
            sheet_id = url.split("/d/")[1].split("/edit")[0]
        else:
            raise ValueError("Invalid Google Sheet URL format.")

        export_url = (
            f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
        )

        # Download the Google Sheet as an Excel file
        response = requests.get(export_url)
        response.raise_for_status()  # Ensure the request was successful

        # Save the Excel file to a temporary file
        with tempfile.NamedTemporaryFile(suffix=".xlsx", delete=False) as temp_file:
            temp_file.write(response.content)
            temp_filename = temp_file.name

        # First create the ScenarioList with default column names
        scenario_list = cls.from_excel(temp_filename, sheet_name=sheet_name)

        # If column_names is provided, create a new ScenarioList with the specified names
        if column_names is not None:
            if len(column_names) != len(scenario_list[0].keys()):
                raise ValueError(
                    f"Number of provided column names ({len(column_names)}) "
                    f"does not match number of columns in sheet ({len(scenario_list[0].keys())})"
                )
            
            # Create a codebook mapping original keys to new names
            original_keys = list(scenario_list[0].keys())
            codebook = dict(zip(original_keys, column_names))
            
            # Return new ScenarioList with renamed columns
            return scenario_list.rename(codebook)
        else:
            return scenario_list

    @classmethod
    def from_delimited_file(
        cls, source: Union[str, "ParseResult"], delimiter: str = ","
    ) -> ScenarioList:
        """Create a ScenarioList from a delimited file (CSV/TSV) or URL."""
        import requests
        from edsl.scenarios.Scenario import Scenario
        from urllib.parse import urlparse
        from urllib.parse import ParseResult

        headers = {
            "Accept": "text/csv,application/csv,text/plain",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36",
        }

        def is_url(source):
            try:
                result = urlparse(source)
                return all([result.scheme, result.netloc])
            except ValueError:
                return False

        try:
            if isinstance(source, str) and is_url(source):
                response = requests.get(source, headers=headers)
                response.raise_for_status()
                file_obj = StringIO(response.text)
            elif isinstance(source, ParseResult):
                response = requests.get(source.geturl(), headers=headers)
                response.raise_for_status()
                file_obj = StringIO(response.text)
            else:
                file_obj = open(source, "r")

            reader = csv.reader(file_obj, delimiter=delimiter)
            header = next(reader)
            observations = [Scenario(dict(zip(header, row))) for row in reader]

        finally:
            file_obj.close()

        return cls(observations)

    # Convenience methods for specific file types
    @classmethod
    def from_csv(cls, source: Union[str, "ParseResult"]) -> ScenarioList:
        """Create a ScenarioList from a CSV file or URL."""
        return cls.from_delimited_file(source, delimiter=",")

    def left_join(self, other: ScenarioList, by: Union[str, list[str]]) -> ScenarioList:
        """Perform a left join with another ScenarioList, following SQL join semantics.

        Args:
            other: The ScenarioList to join with
            by: String or list of strings representing the key(s) to join on. Cannot be empty.

        >>> s1 = ScenarioList([Scenario({'name': 'Alice', 'age': 30}), Scenario({'name': 'Bob', 'age': 25})])
        >>> s2 = ScenarioList([Scenario({'name': 'Alice', 'location': 'New York'}), Scenario({'name': 'Charlie', 'location': 'Los Angeles'})])
        >>> s3 = s1.left_join(s2, 'name')
        >>> s3 == ScenarioList([Scenario({'age': 30, 'location': 'New York', 'name': 'Alice'}), Scenario({'age': 25, 'location': None, 'name': 'Bob'})])
        True
        """
        from edsl.scenarios.scenario_join import ScenarioJoin

        sj = ScenarioJoin(self, other)
        return sj.left_join(by)

    @classmethod
    def from_tsv(cls, source: Union[str, "ParseResult"]) -> ScenarioList:
        """Create a ScenarioList from a TSV file or URL."""
        return cls.from_delimited_file(source, delimiter="\t")

    def to_dict(self, sort: bool = False, add_edsl_version: bool = True) -> dict:
        """
        >>> s = ScenarioList([Scenario({'food': 'wood chips'}), Scenario({'food': 'wood-fired pizza'})])
        >>> s.to_dict()
        {'scenarios': [{'food': 'wood chips', 'edsl_version': '...', 'edsl_class_name': 'Scenario'}, {'food': 'wood-fired pizza', 'edsl_version': '...', 'edsl_class_name': 'Scenario'}], 'edsl_version': '...', 'edsl_class_name': 'ScenarioList'}

        """
        if sort:
            data = sorted(self, key=lambda x: hash(x))
        else:
            data = self
        d = {"scenarios": [s.to_dict(add_edsl_version=add_edsl_version) for s in data]}

        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = self.__class__.__name__
        return d

    def to(self, survey: Union["Survey", "QuestionBase"]) -> "Jobs":
        """Create a Jobs object from a ScenarioList and a Survey object.

        :param survey: The Survey object to use for the Jobs object.

        Example:
        >>> from edsl import Survey
        >>> from edsl.jobs.Jobs import Jobs
        >>> from edsl import ScenarioList
        >>> isinstance(ScenarioList.example().to(Survey.example()), Jobs)
        True
        """
        from edsl.surveys.Survey import Survey
        from edsl.questions.QuestionBase import QuestionBase
        from edsl.jobs.Jobs import Jobs

        if isinstance(survey, QuestionBase):
            return Survey([survey]).by(self)
        else:
            return survey.by(self)

    @classmethod
    def gen(cls, scenario_dicts_list: List[dict]) -> ScenarioList:
        """Create a `ScenarioList` from a list of dictionaries.

        Example:

        >>> ScenarioList.gen([{'name': 'Alice'}, {'name': 'Bob'}])
        ScenarioList([Scenario({'name': 'Alice'}), Scenario({'name': 'Bob'})])

        """
        from edsl.scenarios.Scenario import Scenario

        return cls([Scenario(s) for s in scenario_dicts_list])

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data) -> ScenarioList:
        """Create a `ScenarioList` from a dictionary."""
        from edsl.scenarios.Scenario import Scenario

        return cls([Scenario.from_dict(s) for s in data["scenarios"]])

    @classmethod
    def from_nested_dict(cls, data: dict) -> ScenarioList:
        """Create a `ScenarioList` from a nested dictionary.

        >>> data = {"headline": ["Armistice Signed, War Over: Celebrations Erupt Across City"], "date": ["1918-11-11"], "author": ["Jane Smith"]}
        >>> ScenarioList.from_nested_dict(data)
        ScenarioList([Scenario({'headline': 'Armistice Signed, War Over: Celebrations Erupt Across City', 'date': '1918-11-11', 'author': 'Jane Smith'})])

        """
        length_of_first_list = len(next(iter(data.values())))
        s = ScenarioList.create_empty_scenario_list(n=length_of_first_list)

        if any(len(v) != length_of_first_list for v in data.values()):
            raise ValueError(
                "All lists in the dictionary must be of the same length.",
            )
        for key, list_of_values in data.items():
            s = s.add_list(key, list_of_values)
        return s

    def code(self) -> str:
        """Create the Python code representation of a survey."""
        header_lines = [
            "from edsl.scenarios.Scenario import Scenario",
            "from edsl.scenarios.ScenarioList import ScenarioList",
        ]
        lines = ["\n".join(header_lines)]
        names = []
        for index, scenario in enumerate(self):
            lines.append(f"scenario_{index} = " + repr(scenario))
            names.append(f"scenario_{index}")
        lines.append(f"scenarios = ScenarioList([{', '.join(names)}])")
        return lines

    @classmethod
    def example(cls, randomize: bool = False) -> ScenarioList:
        """
        Return an example ScenarioList instance.

        :params randomize: If True, use Scenario's randomize method to randomize the values.
        """
        return cls([Scenario.example(randomize), Scenario.example(randomize)])

    # def rich_print(self) -> None:
    #     """Display an object as a table."""
    #     from rich.table import Table

    #     table = Table(title="ScenarioList")
    #     table.add_column("Index", style="bold")
    #     table.add_column("Scenario")
    #     for i, s in enumerate(self):
    #         table.add_row(str(i), s.rich_print())
    #     return table

    def __getitem__(self, key: Union[int, slice]) -> Any:
        """Return the item at the given index.

        Example:
        >>> s = ScenarioList([Scenario({'age': 22, 'hair': 'brown', 'height': 5.5}), Scenario({'age': 22, 'hair': 'brown', 'height': 5.5})])
        >>> s[0]
        Scenario({'age': 22, 'hair': 'brown', 'height': 5.5})

        >>> s[:1]
        ScenarioList([Scenario({'age': 22, 'hair': 'brown', 'height': 5.5})])

        """
        if isinstance(key, slice):
            return ScenarioList(super().__getitem__(key))
        elif isinstance(key, int):
            return super().__getitem__(key)
        else:
            return self.to_dict(add_edsl_version=False)[key]

    def to_agent_list(self):
        """Convert the ScenarioList to an AgentList.

        Example:

        >>> s = ScenarioList([Scenario({'age': 22, 'hair': 'brown', 'height': 5.5}), Scenario({'age': 22, 'hair': 'brown', 'height': 5.5})])
        >>> s.to_agent_list()
        AgentList([Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5}), Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})])
        """
        from edsl.agents.AgentList import AgentList
        from edsl.agents.Agent import Agent
        import warnings

        agents = []
        for scenario in self:
            new_scenario = scenario.copy().data
            if "name" in new_scenario:
                name = new_scenario.pop("name")
                proposed_agent_name = "agent_name"
                while proposed_agent_name not in new_scenario:
                    proposed_agent_name += "_"
                warnings.warn(
                    f"The 'name' field is reserved for the agent's name---putting this value in {proposed_agent_name}"
                )
                new_scenario[proposed_agent_name] = name
                new_agent = Agent(traits=new_scenario, name=name)
            if "agent_parameters" in new_scenario:
                agent_parameters = new_scenario.pop("agent_parameters")
                instruction = agent_parameters.get("instruction", None)
                name = agent_parameters.get("name", None)
                new_agent = Agent(
                    traits=new_scenario, name=name, instruction=instruction
                )
            else:
                new_agent = Agent(traits=new_scenario)

            agents.append(new_agent)

        return AgentList(agents)

    def chunk(
        self,
        field,
        num_words: Optional[int] = None,
        num_lines: Optional[int] = None,
        include_original=False,
        hash_original=False,
    ) -> "ScenarioList":
        """Chunk the scenarios based on a field.

        Example:

        >>> s = ScenarioList([Scenario({'text': 'The quick brown fox jumps over the lazy dog.'})])
        >>> s.chunk('text', num_words=3)
        ScenarioList([Scenario({'text': 'The quick brown', 'text_chunk': 0, 'text_char_count': 15, 'text_word_count': 3}), Scenario({'text': 'fox jumps over', 'text_chunk': 1, 'text_char_count': 14, 'text_word_count': 3}), Scenario({'text': 'the lazy dog.', 'text_chunk': 2, 'text_char_count': 13, 'text_word_count': 3})])
        """
        new_scenarios = []
        for scenario in self:
            replacement_scenarios = scenario.chunk(
                field,
                num_words=num_words,
                num_lines=num_lines,
                include_original=include_original,
                hash_original=hash_original,
            )
            new_scenarios.extend(replacement_scenarios)
        return ScenarioList(new_scenarios)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
from abc import ABCMeta

from edsl.enums import QuestionType
from edsl.exceptions.questions import QuestionMissingTypeError, QuestionBadTypeError

import inspect


class RegisterQuestionsMeta(ABCMeta):
    """Metaclass to register output elements in a registry i.e., those that have a parent."""

    _registry = {}  # Initialize the registry as a dictionary

    def __init__(cls, name, bases, dct):
        """Initialize the class and adds it to the registry if it's not the base class."""
        super(RegisterQuestionsMeta, cls).__init__(name, bases, dct)
        if (
            name != "QuestionBase"
            and name != "QuestionFunctional"
            and name != "QuestionAddTwoNumbers"
        ):
            ## Enforce that all questions have a question_type class attribute
            ## and it comes from our enum of valid question types.
            required_attributes = [
                "question_type",
                "_response_model",
                "response_validator_class",
            ]
            for attr in required_attributes:
                if not hasattr(cls, attr):
                    raise QuestionMissingTypeError(
                        f"Question must have a {attr} class attribute"
                    )

            init_signature = inspect.signature(cls.__init__)
            init_params = [p for p in init_signature.parameters if p != "self"]
            required_params = [
                "question_presentation",
                "answering_instructions",
                "question_name",
                "question_text",
            ]
            for param in required_params:
                if param not in init_params:
                    raise QuestionBadTypeError(
                        f"Question type {name} must have a question_presentation parameter in its __init__ method"
                    )

        if name != "QuestionBase":
            RegisterQuestionsMeta._registry[name] = cls

    @classmethod
    def get_registered_classes(cls):
        """Return the registry of registered classes."""
        return cls._registry

    @classmethod
    def question_types_to_classes(
        cls,
    ):
        """Return a dictionary of question types to classes."""
        d = {}
        for classname, cls in cls._registry.items():
            if hasattr(cls, "question_type"):
                d[cls.question_type] = cls
            else:
                raise Exception(
                    f"Class {classname} does not have a question_type class attribute"
                )
        return d



"""Mixin class for ggplot2 plotting."""

import subprocess
import tempfile
from typing import Optional


class GGPlot:
    """A class to handle ggplot2 plot display and saving."""
    
    def __init__(self, r_code: str, width: float = 6, height: float = 4):
        """Initialize with R code and dimensions."""
        self.r_code = r_code
        self.width = width
        self.height = height
        self._svg_data = None
        self._saved = False  # Track if the plot was saved
        
    def _execute_r_code(self, save_command: str = ""):
        """Execute R code with optional save command."""
        full_r_code = self.r_code + save_command
        
        result = subprocess.run(
            ["Rscript", "-"],
            input=full_r_code,
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if result.returncode != 0:
            if result.returncode == 127:
                raise RuntimeError(
                    "Rscript is probably not installed. Please install R from https://cran.r-project.org/"
                )
            else:
                raise RuntimeError(
                    f"An error occurred while running Rscript: {result.stderr}"
                )

        if result.stderr:
            print("Error in R script:", result.stderr)
            
        return result
        
    def save(self, filename: str):
        """Save the plot to a file."""
        format = filename.split('.')[-1].lower()
        if format not in ['svg', 'png']:
            raise ValueError("Only 'svg' and 'png' formats are supported")
            
        save_command = f'\nggsave("{filename}", plot = last_plot(), width = {self.width}, height = {self.height}, device = "{format}")'
        self._execute_r_code(save_command)
        
        self._saved = True
        print(f"File saved to: {filename}")
        return None  # Return None instead of self
        
    def _repr_html_(self):
        """Display the plot in a Jupyter notebook."""
        # Don't display if the plot was saved
        if self._saved:
            return None
            
        import tempfile
        
        # Generate SVG if we haven't already
        if self._svg_data is None:
            # Create temporary SVG file
            with tempfile.NamedTemporaryFile(suffix='.svg') as tmp:
                save_command = f'\nggsave("{tmp.name}", plot = last_plot(), width = {self.width}, height = {self.height}, device = "svg")'
                self._execute_r_code(save_command)
                with open(tmp.name, 'r') as f:
                    self._svg_data = f.read()
                    
        return self._svg_data

class GGPlotMethod:

    def __init__(self, results: 'Results'):
        self.results = results
    
    """Mixin class for ggplot2 plotting."""

    def ggplot2(
        self,
        ggplot_code: str,
        shape="wide",
        sql: str = None,
        remove_prefix: bool = True,
        debug: bool = False,
        height=4,
        width=6,
        factor_orders: Optional[dict] = None,
    ):
        """Create a ggplot2 plot from a DataFrame.

        Returns a GGPlot object that can be displayed in a notebook or saved to a file.

        :param ggplot_code: The ggplot2 code to execute.
        :param shape: The shape of the data in the DataFrame (wide or long).
        :param sql: The SQL query to execute beforehand to manipulate the data.
        :param remove_prefix: Whether to remove the prefix from the column names.
        :param debug: Whether to print the R code instead of executing it.
        :param height: The height of the plot in inches.
        :param width: The width of the plot in inches.
        :param factor_orders: A dictionary of factor columns and their order.
        """
        if sql == None:
            sql = "select * from self"

        if shape == "long":
            df = self.results.sql(sql, shape="long")
        elif shape == "wide":
            df = self.results.sql(sql, remove_prefix=remove_prefix)

        # Convert DataFrame to CSV format
        csv_data = df.to_csv().text

        # Embed the CSV data within the R script
        csv_data_escaped = csv_data.replace("\n", "\\n").replace("'", "\\'")
        read_csv_code = f"self <- read.csv(text = '{csv_data_escaped}', sep = ',')\n"

        if factor_orders is not None:
            for factor, order in factor_orders.items():
                level_string = ", ".join([f'"{x}"' for x in order])
                read_csv_code += (
                    f"self${factor} <- factor(self${factor}, levels=c({level_string}))"
                )
                read_csv_code += "\n"

        # Load ggplot2 library and combine all R script parts
        full_r_code = "library(ggplot2)\n" + read_csv_code + ggplot_code

        if debug:
            print(full_r_code)
            return

        return GGPlot(full_r_code, width=width, height=height)

    def _display_plot(self, filename: str, width: float, height: float):
        """Display the plot in the notebook or open in system viewer if running from terminal."""
        try:
            # Try to import IPython-related modules
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg
            from IPython import get_ipython

            # Check if we're in a notebook environment
            if get_ipython() is not None:
                if filename.endswith(".png"):
                    img = mpimg.imread(filename)
                    plt.figure(figsize=(width, height))
                    plt.imshow(img)
                    plt.axis("off")
                    plt.show()
                elif filename.endswith(".svg"):
                    from IPython.display import SVG, display
                    display(SVG(filename=filename))
                else:
                    print("Unsupported file format. Please provide a PNG or SVG file.")
                return

        except ImportError:
            pass

        # If we're not in a notebook or imports failed, open with system viewer
        import platform
        import os
        
        system = platform.system()
        if system == 'Darwin':       # macOS
            if filename.endswith('.svg'):
                subprocess.run(['open', '-a', 'Preview', filename])
            else:
                subprocess.run(['open', filename])
        elif system == 'Linux':
            subprocess.run(['xdg-open', filename])
        elif system == 'Windows':
            os.startfile(filename)
        else:
            print(f"File saved to: {filename}")



from textwrap import dedent
from dataclasses import dataclass
from collections import defaultdict

from typing import List, Dict, Union

from edsl.auto.StageBase import StageBase
from edsl.auto.StageBase import FlowDataBase

from edsl.auto.StageQuestions import StageQuestions

from edsl.questions import QuestionMultipleChoice, QuestionList
from edsl.scenarios import Scenario
from edsl import Model
from edsl.auto.utilities import gen_pipeline


question_purpose = {
    "multiple_choice": "When options are known and limited",
    "free_text": "When we are asking an open-ended question",
    "checkbox": "When multiple options can be selected e.g., have you heard of the following products:",
    "numerical": "When the answer is a single numerical value e.g., a float",
    "linear_scale": "When options are text like multiple choice, but can be ordered e.g., daily, weekly, monthly, etc.",
    "yes_no": "When the question can be fully answered with either a yes or a no",
}


class StageLabelQuestions(StageBase):
    input = StageQuestions.output

    @dataclass
    class Output(FlowDataBase):
        questions: List[str]
        types: List[str]
        options: Dict[str, List[str]]
        option_labels: Dict[str, Union[List[str], None]]

    output = Output

    def handle_data(self, data):
        """
        Labels each edsl question type. This is then used later to instantiate the questions
        """
        m = Model()
        label_questions_scenarios = [
            Scenario({"question": q, "question_purpose": question_purpose})
            for q in data.questions
        ]
        q_type = QuestionMultipleChoice(
            question_text=dedent(
                """\
        Consider this question: "{{ question }}"
        The question options and purpose are: {{ question_purpose }}
        Please avoid free text questions much as possible. 
        If it could be a multiple choice, use that type.
        What type of question should this be to make for an informative survey?"""
            ),
            question_options=list(question_purpose.keys()),
            question_name="question_type",
        )
        ## If it is a linear scale, multiple choice or checkbox question, we need to know the options
        option_questions = [
            "multiple_choice",
            "linear_scale",
            "checkbox",
        ]
        q_options_mc = QuestionList(
            question_text=dedent(
                """\
            Consider this question: "{{ question }}"
            What options should this question have?"""
            ),
            question_name="mc_options",
        )
        survey = q_type.add_question(q_options_mc).add_stop_rule(
            "question_type", f"question_type not in {option_questions}"
        )
        type_results = survey.by(label_questions_scenarios).by(m).run()
        type_results.select("question", "question_type", "mc_options").print()

        # breakpoint()

        question_types = type_results.select("question_type").to_list()
        options = type_results.select("mc_options").to_list()
        # question_types, options = type_results.select(
        #    "question_type", "mc_options"
        # ).to_list()

        type_results.select("question", "question_type", "mc_options").print()

        # if the question is a yes/no question, we need to set the options to be yes/no
        types_to_questions = defaultdict(list)
        for question_type, question in zip(question_types, data.questions):
            types_to_questions[question_type].append(question)

        questions_to_options = dict(zip(data.questions, options))
        question_to_option_labels = dict(
            zip(data.questions, len(data.questions) * [None])
        )
        for question in types_to_questions.get("yes_no", []):
            questions_to_options[question] = ["Yes", "No"]

        for question in types_to_questions.get("linear_scale", []):
            options = questions_to_options[question]
            questions_to_options[question] = list(range(len(options)))
            question_to_option_labels[question] = options

        return self.output(
            questions=data.questions,
            types=question_types,
            options=list(questions_to_options.values()),
            option_labels=list(question_to_option_labels.values()),
        )


if __name__ == "__main__":
    pipeline = gen_pipeline([StageQuestions, StageLabelQuestions])

    results = pipeline.process(
        pipeline.input(
            overall_question="What are some factors that could determine whether someone likes ice cream?"
        )
    )

    print(results.options)



import difflib
import json
from typing import Any, Dict, Tuple
from collections import UserList
import inspect


class BaseDiffCollection(UserList):
    def __init__(self, diffs=None):
        if diffs is None:
            diffs = []
        super().__init__(diffs)

    def apply(self, obj: Any):
        for diff in self:
            obj = diff.apply(obj)
        return obj

    def add_diff(self, diff) -> "BaseDiffCollection":
        self.append(diff)
        return self


class DummyObject:
    def __init__(self, object_dict):
        self.object_dict = object_dict

    def to_dict(self):
        return self.object_dict


class BaseDiff:
    def __init__(
        self, obj1: Any, obj2: Any, added=None, removed=None, modified=None, level=0
    ):
        self.level = level

        self.obj1 = obj1
        self.obj2 = obj2

        if "sort" in inspect.signature(obj1.to_dict).parameters:
            self._dict1 = obj1.to_dict(sort=True)
            self._dict2 = obj2.to_dict(sort=True)
        else:
            self._dict1 = obj1.to_dict()
            self._dict2 = obj2.to_dict()
        self._obj_class = type(obj1)

        self.added = added
        self.removed = removed
        self.modified = modified

    def __bool__(self):
        return bool(self.added or self.removed or self.modified)

    @property
    def added(self):
        if self._added is None:
            self._added = self._find_added()
        return self._added

    def __add__(self, other):
        return self.apply(other)

    @added.setter
    def added(self, value):
        self._added = value if value is not None else self._find_added()

    @property
    def removed(self):
        if self._removed is None:
            self._removed = self._find_removed()
        return self._removed

    @removed.setter
    def removed(self, value):
        self._removed = value if value is not None else self._find_removed()

    @property
    def modified(self):
        if self._modified is None:
            self._modified = self._find_modified()
        return self._modified

    @modified.setter
    def modified(self, value):
        self._modified = value if value is not None else self._find_modified()

    def _find_added(self) -> Dict[Any, Any]:
        return {k: self._dict2[k] for k in self._dict2 if k not in self._dict1}

    def _find_removed(self) -> Dict[Any, Any]:
        return {k: self._dict1[k] for k in self._dict1 if k not in self._dict2}

    def _find_modified(self) -> Dict[Any, Tuple[Any, Any, str]]:
        modified = {}
        for k in self._dict1:
            if k in self._dict2 and self._dict1[k] != self._dict2[k]:
                if isinstance(self._dict1[k], str) and isinstance(self._dict2[k], str):
                    diff = self._diff_strings(self._dict1[k], self._dict2[k])
                    modified[k] = (self._dict1[k], self._dict2[k], diff)
                elif isinstance(self._dict1[k], dict) and isinstance(
                    self._dict2[k], dict
                ):
                    diff = self._diff_dicts(self._dict1[k], self._dict2[k])
                    modified[k] = (self._dict1[k], self._dict2[k], diff)
                elif isinstance(self._dict1[k], list) and isinstance(
                    self._dict2[k], list
                ):
                    d1 = dict(zip(range(len(self._dict1[k])), self._dict1[k]))
                    d2 = dict(zip(range(len(self._dict2[k])), self._dict2[k]))
                    diff = BaseDiff(
                        DummyObject(d1), DummyObject(d2), level=self.level + 1
                    )
                    modified[k] = (self._dict1[k], self._dict2[k], diff)
                else:
                    modified[k] = (self._dict1[k], self._dict2[k], "")
        return modified

    @staticmethod
    def is_json(string_that_could_be_json: str) -> bool:
        try:
            json.loads(string_that_could_be_json)
            return True
        except json.JSONDecodeError:
            return False

    def _diff_dicts(self, dict1: Dict[str, Any], dict2: Dict[str, Any]) -> str:
        diff = BaseDiff(DummyObject(dict1), DummyObject(dict2), level=self.level + 1)
        return diff

    def _diff_strings(self, str1: str, str2: str) -> str:
        if self.is_json(str1) and self.is_json(str2):
            diff = self._diff_dicts(json.loads(str1), json.loads(str2))
            return diff
        diff = difflib.ndiff(str1.splitlines(), str2.splitlines())
        return diff

    def apply(self, obj: Any):
        """Apply the diff to the object."""

        new_obj_dict = obj.to_dict()
        for k, v in self.added.items():
            new_obj_dict[k] = v
        for k in self.removed.keys():
            del new_obj_dict[k]
        for k, (v1, v2, diff) in self.modified.items():
            new_obj_dict[k] = v2

        return obj.from_dict(new_obj_dict)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "added": self.added,
            "removed": self.removed,
            "modified": self.modified,
            "obj1": self._dict1,
            "obj2": self._dict2,
            "obj_class": self._obj_class.__name__,
            "level": self.level,
        }

    @classmethod
    def from_dict(cls, diff_dict: Dict[str, Any], obj1: Any, obj2: Any):
        return cls(
            obj1=obj1,
            obj2=obj2,
            added=diff_dict["added"],
            removed=diff_dict["removed"],
            modified=diff_dict["modified"],
            level=diff_dict["level"],
        )

    class Results(UserList):
        def __init__(self, prepend=" ", level=0):
            super().__init__()
            self.prepend = prepend
            self.level = level

        def append(self, item):
            super().append(self.prepend * self.level + item)

    def __str__(self):
        prepend = " "
        result = self.Results(level=self.level, prepend="\t")
        if self.added:
            result.append("Added keys and values:")
            for k, v in self.added.items():
                result.append(prepend + f"  {k}: {v}")
        if self.removed:
            result.append("Removed keys and values:")
            for k, v in self.removed.items():
                result.append(f"  {k}: {v}")
        if self.modified:
            result.append("Modified keys and values:")
            for k, (v1, v2, diff) in self.modified.items():
                result.append(f"Key: {k}:")
                result.append(f"    Old value: {v1}")
                result.append(f"    New value: {v2}")
                if diff:
                    result.append(f"    Diff:")
                    try:
                        for line in diff:
                            result.append(f"      {line}")
                    except:
                        result.append(f"      {diff}")
        return "\n".join(result)

    def __repr__(self):
        return (
            f"BaseDiff(obj1={self.obj1!r}, obj2={self.obj2!r}, added={self.added!r}, "
            f"removed={self.removed!r}, modified={self.modified!r})"
        )

    def add_diff(self, diff) -> "BaseDiffCollection":
        return BaseDiffCollection([self, diff])


if __name__ == "__main__":
    from edsl import Question

    q_ft = Question.example("free_text")
    q_mc = Question.example("multiple_choice")

    diff1 = q_ft - q_mc
    assert q_ft == q_mc + diff1
    assert q_ft == diff1.apply(q_mc)
    # new_q_mc = diff1.apply(q_ft)
    # assert new_q_mc == q_mc

    # new_q_mc = q_ft + diff1
    # assert new_q_mc == q_mc

    # new_q_mc = diff1 + q_ft
    # assert new_q_mc == q_mc

    # ## Test chain of diffs
    q0 = Question.example("free_text")
    q1 = q0.copy()
    q1.question_text = "Why is Buzzard's Bay so named?"
    diff1 = q1 - q0
    q2 = q1.copy()
    q2.question_name = "buzzard_bay"
    diff2 = q2 - q1

    diff_chain = diff1.add_diff(diff2)

    new_q2 = diff_chain.apply(q0)
    assert new_q2 == q2

    new_q2 = diff_chain + q0
    assert new_q2 == q2

    # new_diffs = diff1.add_diff(diff1).add_diff(diff1)
    # assert len(new_diffs) == 3

    # q0 = Question.example("free_text")
    # q1 = Question.example("free_text")
    # q1.question_text = "Why is Buzzard's Bay so named?"
    # q2 = q1.copy()



from edsl import Agent, AgentList, QuestionFreeText
from edsl import Cache
from edsl.conversation.Conversation import Conversation, ConversationList

a1 = Agent(
    name="Alice",
    traits={
        "motivation": """
    You are Alice. You want to buy a car. You are talking to Bob, a car salesman.
    It is very important to you that the steering wheel not whiff out of the window while you are driving.
    Also, the car can have no space for mother-in-law.
    You brought your brother-in-law along, Paul, who you have an antagonistic relationship with.
    """
    },
)
a2 = Agent(
    name="Bob",
    traits={
        "motivation": "You are Bob, a car salesman. You are talking to Alice, who wants to buy a car."
    },
)
a3 = Agent(
    name="Paul",
    traits={
        "motivation": "You are Paul, you are Alice's brother. You think her concerns are foolish and you are critical of her."
    },
)

c1 = Conversation(agent_list=AgentList([a1, a3, a2]), max_turns=5, verbose=True)
c2 = Conversation(agent_list=AgentList([a1, a2]), max_turns=5, verbose=True)

# c = Cache.load("car_talk.json.gz")
c = Cache()
# breakpoint()
combo = ConversationList([c1, c2], cache=c)
combo.run()
results = combo.to_results()
results.select("conversation_index", "index", "agent_name", "dialogue").print(
    format="rich"
)

q = QuestionFreeText(
    question_text="""This was a conversation about buying a car: {{ transcript }}. 
                    Was a brand or style of car mentioned? If so, what was it?
                    """,
    question_name="car_brand",
)

from edsl import QuestionList

q_actors = QuestionList(
    question_text="""This was a conversation about buying a car: {{ transcript }}. 
                    Who were the actors in the conversation?
                    """,
    question_name="actors",
)

transcript_analysis = q.add_question(q_actors).by(combo.summarize()).run()
transcript_analysis.select("car_brand", "actors").print(format="rich")



import textwrap
from random import random
from typing import Optional, TYPE_CHECKING, List, Callable

from edsl.utilities.PrettyList import PrettyList
from edsl.config import CONFIG

from edsl.inference_services.InferenceServicesCollection import (
    InferenceServicesCollection,
)
from edsl.inference_services.data_structures import AvailableModels
from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.enums import InferenceServiceLiteral
from edsl.exceptions.inference_services import InferenceServiceError

if TYPE_CHECKING:
    from edsl.results.Dataset import Dataset


def get_model_class(
    model_name,
    registry: Optional[InferenceServicesCollection] = None,
    service_name: Optional[InferenceServiceLiteral] = None,
):
    from edsl.inference_services.registry import default

    registry = registry or default
    try:
        factory = registry.create_model_factory(model_name, service_name=service_name)
        return factory
    except (InferenceServiceError, Exception) as e:
        return Model._handle_model_error(model_name, e)


class Meta(type):
    def __repr__(cls):
        return textwrap.dedent(
            f"""\
        Available models: {cls.available()}
        
        To create an instance, you can do: 
        >>> m = Model('gpt-4-1106-preview', temperature=0.5, ...)
        
        To get the default model, you can leave out the model name. 
        To see the available models, you can do:
        >>> Model.available()

        Or to see the models for a specific service, you can do:
        >>> Model.available(service='openai')
        """
        )


class Model(metaclass=Meta):
    default_model = CONFIG.get("EDSL_DEFAULT_MODEL")
    _registry: InferenceServicesCollection = None  # Class-level registry storage

    @classmethod
    def get_registry(cls) -> InferenceServicesCollection:
        """Get the current registry or initialize with default if None"""
        if cls._registry is None:
            from edsl.inference_services.registry import default

            cls._registry = default
        return cls._registry

    @classmethod
    def set_registry(cls, registry: InferenceServicesCollection) -> None:
        """Set a new registry"""
        cls._registry = registry

    @classmethod
    def _handle_model_error(cls, model_name: str, error: Exception):
        """Handle errors from model creation and execution with notebook-aware behavior."""
        if isinstance(error, InferenceServiceError):
            services = [s._inference_service_ for s in cls.get_registry().services]
            message = (
                f"Model '{model_name}' not found in any services.\n"
                "It is likely that our registry is just out of date.\n"
                "Simply adding the service name to your model call should fix this.\n"
                f"Available services are: {services}\n"
                f"To specify a model with a service, use:\n"
                f'Model("{model_name}", service_name="<service_name>")'
            )
        else:
            message = f"An error occurred: {str(error)}"

        # Check if we're in a notebook environment
        try:
            get_ipython()
            print(message)
            return None
        except NameError:
            # Not in a notebook, raise the exception
            if isinstance(error, InferenceServiceError):
                raise InferenceServiceError(message)
            raise error

    def __new__(
        cls,
        model_name: Optional[str] = None,
        service_name: Optional[InferenceServiceLiteral] = None,
        registry: Optional[InferenceServicesCollection] = None,
        *args,
        **kwargs,
    ):
        """Instantiate a new language model.
        >>> Model()
        Model(...)
        """
        # Map index to the respective subclass
        if model_name is None:
            model_name = cls.default_model

        if registry is not None:
            cls.set_registry(registry)

        if isinstance(model_name, int):  # can refer to a model by index
            model_name = cls.available(name_only=True)[model_name]

        try:
            factory = cls.get_registry().create_model_factory(
                model_name, service_name=service_name
            )
            return factory(*args, **kwargs)
        except (InferenceServiceError, Exception) as e:
            return cls._handle_model_error(model_name, e)

    @classmethod
    def add_model(cls, service_name, model_name) -> None:
        cls.get_registry().add_model(service_name, model_name)

    @classmethod
    def service_classes(cls) -> List["InferenceServiceABC"]:
        """Returns a list of service classes.

        >>> Model.service_classes()
        [...]
        """
        return [r for r in cls.services()]

    @classmethod
    def services(cls, name_only: bool = False) -> List[str]:
        """Returns a list of services excluding 'test', sorted alphabetically.

        >>> Model.services()
        [...]
        """
        return PrettyList(
            sorted(
                [
                    [r._inference_service_]
                    for r in cls.get_registry().services
                    if r._inference_service_.lower() != "test"
                ]
            ),
            columns=["Service Name"],
        )

    @classmethod
    def services_with_local_keys(cls) -> set:
        """Returns a list of services for which the user has local keys."""
        return set(cls.key_info().select("service").to_list())

    @classmethod
    def key_info(cls, obscure_api_key: bool = True) -> "Dataset":
        """Returns a dataset of local key information."""
        from edsl.language_models.key_management.KeyLookupCollection import (
            KeyLookupCollection,
        )
        from edsl.scenarios import Scenario, ScenarioList

        klc = KeyLookupCollection()
        klc.add_key_lookup(fetch_order=None)
        sl = ScenarioList()
        for service, entry in list(klc.data.values())[0].items():
            sl.append(Scenario({"service": service} | entry.to_dict()))
        if obscure_api_key:
            for service in sl:
                service["api_token"] = (
                    service["api_token"][:4] + "..." + service["api_token"][-4:]
                )
        return sl.to_dataset()

    @classmethod
    def search_models(cls, search_term: str):
        return cls.available(search_term=search_term)

    @classmethod
    def all_known_models(cls) -> "AvailableModels":
        return cls.get_registry().available()

    @classmethod
    def available_with_local_keys(cls):
        services_with_local_keys = set(cls.key_info().select("service").to_list())
        return [
            m
            for m in cls.get_registry().available()
            if m.service_name in services_with_local_keys
        ]

    @classmethod
    def available(
        cls,
        search_term: str = None,
        name_only: bool = False,
        service: Optional[str] = None,
        force_refresh: bool = False,
    ):
        """Get available models

        >>> Model.available()
        [...]
        >>> Model.available(service='openai')
        [...]
        """
        # if search_term is None and service is None:
        #     print("Getting available models...")
        #     print("You have local keys for the following services:")
        #     print(cls.services_with_local_keys())
        #     print("\n")
        #     print("To see models by service, use the 'service' parameter.")
        #     print("E.g., Model.available(service='openai')")
        #     return None

        if service is not None:
            known_services = [x[0] for x in cls.services(name_only=True)]
            if service not in known_services:
                raise ValueError(
                    f"Service {service} not found in available services.",
                    f"Available services are: {known_services}",
                )

        full_list = cls.get_registry().available(
            service=service, force_refresh=force_refresh
        )

        if search_term is None:
            if name_only:
                return PrettyList(
                    [m.model_name for m in full_list],
                    columns=["Model Name"],
                )
            else:
                return PrettyList(
                    [[m.model_name, m.service_name] for m in full_list],
                    columns=["Model Name", "Service Name"],
                )
        else:
            filtered_results = [
                m
                for m in full_list
                if search_term in m.model_name or search_term in m.service_name
            ]
            if name_only:
                return PrettyList(
                    [m.model_name for m in filtered_results],
                    columns=["Model Name"],
                )
            else:
                return PrettyList(
                    [[m.model_name, m.service_name] for m in full_list],
                    columns=["Model Name", "Service Name"],
                )

    @classmethod
    def check_models(cls, verbose=False):
        print("Checking all available models...\n")
        for model in cls.available(name_only=True):
            print(f"Now checking: {model}")
            try:
                m = cls(model)
            except Exception as e:
                print(f"Error creating instance of {model}: {e}")
                continue
            try:
                results = m.hello(verbose)
                if verbose:
                    print(f"Results from model call: {results}")
            except Exception as e:
                print(f"Error calling 'hello' on {model}: {e}")
                continue
            print("OK!")
            print("\n")

    @classmethod
    def check_working_models(
        cls,
        service: Optional[str] = None,
        works_with_text: Optional[bool] = None,
        works_with_images: Optional[bool] = None,
    ) -> list[dict]:
        from edsl.coop import Coop

        c = Coop()
        working_models = c.fetch_working_models()

        if service is not None:
            working_models = [m for m in working_models if m["service"] == service]
        if works_with_text is not None:
            working_models = [
                m for m in working_models if m["works_with_text"] == works_with_text
            ]
        if works_with_images is not None:
            working_models = [
                m for m in working_models if m["works_with_images"] == works_with_images
            ]

        if len(working_models) == 0:
            return []

        else:
            return PrettyList(
                [
                    [
                        m["service"],
                        m["model"],
                        m["works_with_text"],
                        m["works_with_images"],
                        m["usd_per_1M_input_tokens"],
                        m["usd_per_1M_output_tokens"],
                    ]
                    for m in working_models
                ],
                columns=[
                    "Service",
                    "Model",
                    "Works with text",
                    "Works with images",
                    "Price per 1M input tokens (USD)",
                    "Price per 1M output tokens (USD)",
                ],
            )

    @classmethod
    def example(cls, randomize: bool = False) -> "Model":
        """
        Returns an example Model instance.

        >>> Model.example()
        Model(...)

        :param randomize: If True, the temperature is set to a random decimal between 0 and 1.
        """
        temperature = 0.5 if not randomize else round(random(), 2)
        model_name = cls.default_model
        return cls(model_name, temperature=temperature)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    # available = Model.available()
    # m = Model("gpt-4-1106-preview")
    # results = m.execute_model_call("Hello world")
    # print(results)



import re
import textwrap

# from IPython.core.error import UsageError


class AgentListError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ScenarioError(Exception):
    documentation = "https://docs.expectedparrot.com/en/latest/scenarios.html#module-edsl.scenarios.Scenario"

    def __init__(self, message: str):
        self.message = message + "\n" + "Documentation: " + self.documentation
        super().__init__(self.message)

    def __str__(self):
        return self.make_urls_clickable(self.message)

    @staticmethod
    def make_urls_clickable(text):
        url_pattern = r"https?://[^\s]+"
        urls = re.findall(url_pattern, text)
        for url in urls:
            clickable_url = f"\033]8;;{url}\007{url}\033]8;;\007"
            text = text.replace(url, clickable_url)
        return text



import copy
import asyncio

from typing import Union, Type, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from edsl.questions.QuestionBase import QuestionBase
    from edsl.jobs.interviews.Interview import Interview
    from edsl.language_models.key_management.KeyLookup import KeyLookup

from edsl.surveys.base import EndOfSurvey
from edsl.jobs.tasks.task_status_enum import TaskStatus

from edsl.jobs.FetchInvigilator import FetchInvigilator
from edsl.exceptions.language_models import LanguageModelNoResponseError
from edsl.exceptions.questions import QuestionAnswerValidationError
from edsl.data_transfer_models import AgentResponseDict, EDSLResultObjectInput

from edsl.jobs.Answers import Answers


class RetryConfig:
    from edsl.config import CONFIG

    EDSL_BACKOFF_START_SEC = float(CONFIG.get("EDSL_BACKOFF_START_SEC"))
    EDSL_BACKOFF_MAX_SEC = float(CONFIG.get("EDSL_BACKOFF_MAX_SEC"))
    EDSL_MAX_ATTEMPTS = int(CONFIG.get("EDSL_MAX_ATTEMPTS"))


class SkipHandler:

    def __init__(self, interview: "Interview"):
        self.interview = interview
        self.question_index = self.interview.to_index

        self.skip_function: Callable = (
            self.interview.survey.rule_collection.skip_question_before_running
        )

    def should_skip(self, current_question: "QuestionBase") -> bool:
        """Determine if the current question should be skipped."""
        current_question_index = self.question_index[current_question.question_name]
        combined_answers = (
            self.interview.answers
            | self.interview.scenario
            | self.interview.agent["traits"]
        )
        return self.skip_function(current_question_index, combined_answers)

    def cancel_skipped_questions(self, current_question: "QuestionBase") -> None:
        """Cancel the tasks for questions that should be skipped."""
        current_question_index: int = self.interview.to_index[
            current_question.question_name
        ]
        answers = (
            self.interview.answers
            | self.interview.scenario
            | self.interview.agent["traits"]
        )

        # Get the index of the next question, which could also be the end of the survey
        next_question: Union[int, EndOfSurvey] = (
            self.interview.survey.rule_collection.next_question(
                q_now=current_question_index,
                answers=answers,
            )
        )

        def cancel_between(start, end):
            """Cancel the tasks for questions between the start and end indices."""
            for i in range(start, end):
                self.interview.tasks[i].cancel()

        if (next_question_index := next_question.next_q) == EndOfSurvey:
            cancel_between(
                current_question_index + 1, len(self.interview.survey.questions)
            )
            return

        if next_question_index > (current_question_index + 1):
            cancel_between(current_question_index + 1, next_question_index)


class AnswerQuestionFunctionConstructor:
    """Constructs a function that answers a question and records the answer."""

    def __init__(self, interview: "Interview", key_lookup: "KeyLookup"):
        self.interview = interview
        self.key_lookup = key_lookup

        self.had_language_model_no_response_error: bool = False
        self.question_index = self.interview.to_index

        self.skip_function: Callable = (
            self.interview.survey.rule_collection.skip_question_before_running
        )

        self.invigilator_fetcher = FetchInvigilator(
            self.interview, key_lookup=self.key_lookup
        )
        self.skip_handler = SkipHandler(self.interview)

    def _handle_exception(
        self, e: Exception, invigilator: "InvigilatorBase", task=None
    ):
        """Handle an exception that occurred while answering a question."""

        from edsl.jobs.interviews.InterviewExceptionEntry import InterviewExceptionEntry

        answers = copy.copy(
            self.interview.answers
        )  # copy to freeze the answers here for logging
        exception_entry = InterviewExceptionEntry(
            exception=e,
            invigilator=invigilator,
            answers=answers,
        )
        if task:
            task.task_status = TaskStatus.FAILED

        self.interview.exceptions.add(
            invigilator.question.question_name, exception_entry
        )

        if self.interview.raise_validation_errors and isinstance(
            e, QuestionAnswerValidationError
        ):
            raise e

        stop_on_exception = getattr(self.interview, "stop_on_exception", False)
        if stop_on_exception:
            raise e

    def __call__(self):
        return self.answer_question_and_record_task

    async def answer_question_and_record_task(
        self,
        *,
        question: "QuestionBase",
        task=None,
    ) -> "EDSLResultObjectInput":

        from tenacity import (
            retry,
            stop_after_attempt,
            wait_exponential,
            retry_if_exception_type,
            RetryError,
        )

        @retry(
            stop=stop_after_attempt(RetryConfig.EDSL_MAX_ATTEMPTS),
            wait=wait_exponential(
                multiplier=RetryConfig.EDSL_BACKOFF_START_SEC,
                max=RetryConfig.EDSL_BACKOFF_MAX_SEC,
            ),
            retry=retry_if_exception_type(LanguageModelNoResponseError),
            reraise=True,
        )
        async def attempt_answer():
            invigilator = self.invigilator_fetcher(question)

            if self.skip_handler.should_skip(question):
                return invigilator.get_failed_task_result(
                    failure_reason="Question skipped."
                )

            try:
                response: EDSLResultObjectInput = (
                    await invigilator.async_answer_question()
                )
                if response.validated:
                    self.interview.answers.add_answer(
                        response=response, question=question
                    )

                    self.skip_handler.cancel_skipped_questions(question)
                else:
                    if (
                        hasattr(response, "exception_occurred")
                        and response.exception_occurred
                    ):
                        raise response.exception_occurred

            except QuestionAnswerValidationError as e:
                self._handle_exception(e, invigilator, task)
                return invigilator.get_failed_task_result(
                    failure_reason="Question answer validation failed."
                )

            except asyncio.TimeoutError as e:
                self._handle_exception(e, invigilator, task)
                had_language_model_no_response_error = True
                raise LanguageModelNoResponseError(
                    f"Language model timed out for question '{question.question_name}.'"
                )

            except Exception as e:
                self._handle_exception(e, invigilator, task)

            if "response" not in locals():
                had_language_model_no_response_error = True
                raise LanguageModelNoResponseError(
                    f"Language model did not return a response for question '{question.question_name}.'"
                )

            if (
                question.question_name in self.interview.exceptions
                and had_language_model_no_response_error
            ):
                self.interview.exceptions.record_fixed_question(question.question_name)

            return response

        try:
            return await attempt_answer()
        except RetryError as retry_error:
            original_error = retry_error.last_attempt.exception()
            self._handle_exception(
                original_error, self.invigilator_fetcher(question), task
            )
            raise original_error
